import sys
from pathlib import Path
from typing import IO, AnyStr, Union
from textwrap import dedent
import datetime

import numpy as np

from pysprint.core.phase import Phase
from pysprint.config import _get_config_value
from pysprint.utils.exceptions import NotCalculatedException
from pysprint.utils import run_from_ipython

__all__ = ["defaultcallback", "eager_executor"]


PathOrBuffer = Union[str, Path, IO[AnyStr]]


class LogWriter:
    LABELS = ("GD", "GDD", "TOD", "FOD", "QOD", "SOD")

    def __init__(self, file: PathOrBuffer, phase: Phase, verbosity: Union[int, None] = None):
        self.file = file

        if not self.file.endswith((".log", ".txt")):
            self.file += ".log"

        # if os.path.exists(self.file):
        #     warnings.warn(f"File {self.file} exists, opening it in append mode.", PySprintWarning)

        self.phase = phase
        self.verbosity = verbosity or _get_config_value("verbosity")

    def _write_content(self, content):
        with open(self.file, "a") as logfile:
            logfile.write(content)

    def _prepare_content(self):

        if self.phase.coef_array is None:
            raise NotCalculatedException

        precision = _get_config_value("precision")

        iter_num = len(self.phase.x)

        output = dedent(f"""
        ---------------------------------------------------------------------------------------
        Date: {datetime.datetime.now()}

        Datapoints used: {iter_num}

        R^2: {self.phase._get_r_squared():.{precision}f}

        Results:
        """)

        for i, (label, value) in enumerate(zip(self.LABELS, self.phase.coef_array)):
            if value is not None and value != 0:
                output += f"{label} = {value:.{precision}f} fs^{i + 1}\n"

        if self.verbosity > 0:
            with np.printoptions(
                    threshold=sys.maxsize,
                    linewidth=np.inf,
                    precision=precision,
            ):
                output += dedent(f"""
                Values used:
                x: {self.phase.x}

                y: {self.phase.y}
                """)
        return output

    def write(self):
        content = self._prepare_content()
        self._write_content(content)


def defaultcallback(broadcaster, listener=None):
    """
    The default recorder for SPP data.
    """
    if listener is not None:
        listener._container[broadcaster] = broadcaster.emit()


def eager_executor(reference_point=None, order=None, logfile=None, verbosity=None, cleanup=True, **kwds):
    """
    This function can be used as a `callback` function. It will
    constantly monitor for SPP data changes and will re-calculate
    the dispersion on any modification.

    Parameters
    ----------
    reference_point : float
        The reference point to use.
    order : int
        The dispersion order to look for.
    cleanup : bool, optional
        Whether to flush the output and override previous results.
    logfile : string or None, optional
        If given, a logfile will be created each iteration.
    verbosity : int or None, optional
        The verbosity level for the logfile. If 0 or None then it
        won't write the data used for fitting, else it will.
    """
    if not reference_point or not order:
        raise RuntimeError(
            "Reference point and order is required."
        )

    def inner(broadcaster, listener=None):
        defaultcallback(broadcaster, listener)
        if run_from_ipython():
            from IPython.display import clear_output  # noqa
        else:
            import sys
            clear_output = sys.stdout.flush
        try:
            if cleanup:
                clear_output()
            listener.calculate(reference_point, order, **kwds)
            try:
                if logfile:
                    LogWriter(logfile, listener.GD, verbosity).write()
            except NotCalculatedException:
                pass
        except (TypeError, ValueError):
            pass

    return inner



"""
This file contains code from a blog post by Jake VanderPlas.

See at:
https://jakevdp.github.io/blog/2015/02/24/optimizing-python-with-numpy-and-numba/

References:

[1] Dutt A., Rokhlin V. : Fast Fourier Transforms for Nonequispaced Data II,
    Applied and Computational Harmonic Analysis
    Volume 2, Issue 1, January 1995, Pages 85-100
    (1995)

[2] Greengard, Leslie & Lee, June-Yub.: Accelerating the
    Nonuniform Fast Fourier Transform,
    Society for Industrial and Applied Mathematics.
    46. 443-454. 10.1137/S003614450343200X.
    (2004)
"""

import warnings
from inspect import isfunction

import numpy as np

try:
    from numba import jit

    _has_numba = True
except ImportError:
    _has_numba = False

    def jit(func=None, *args, **kwargs):
        """Replace jit if numba is not available."""

        def _jit(f):
            return f

        if isfunction(func):
            return func
        else:
            return _jit


def _grid_params(gl, eps):
    if eps <= 1e-33 or eps >= 1e-1:
        raise ValueError(f"eps = {eps:.0e}, but it must satisfy " "1e-33 < eps < 1e-1.")

    ratio = 2 if eps > 1e-11 else 3
    Msp = int(-np.log(eps) / (np.pi * (ratio - 1) / (ratio - 0.5)) + 0.5)
    Mr = max(ratio * gl, 2 * Msp)
    lambda_ = Msp / (ratio * (ratio - 0.5))
    tau = np.pi * lambda_ / gl ** 2
    return Mr, Msp, tau


@jit(nopython=True)
def _grid(x, c, tau, Msp, ftau, E3):
    Mr = ftau.shape[0]
    hx = 2 * np.pi / Mr

    for j in range(Msp + 1):
        E3[j] = np.exp(-((np.pi * j / Mr) ** 2) / tau)

    for i in range(x.shape[0]):
        xi = x[i] % (2 * np.pi)
        m = 1 + int(xi // hx)
        xi = xi - hx * m
        E1 = np.exp(-0.25 * xi ** 2 / tau)
        E2 = np.exp((xi * np.pi) / (Mr * tau))
        E2mm = 1
        for mm in range(Msp):
            ftau[(m + mm) % Mr] += c[i] * E1 * E2mm * E3[mm]
            E2mm *= E2
            ftau[(m - mm - 1) % Mr] += c[i] * E1 / E2mm * E3[mm + 1]


def _compute_gaussian_grid(x, c, Mr, Msp, tau):
    ftau = np.zeros(Mr, dtype=c.dtype)
    E3 = np.zeros(Msp + 1, dtype=x.dtype)
    _grid(x, c, tau, Msp, ftau, E3)
    return ftau


def nuifft_times(gl, df=1):
    """Compute the time range used in nufft for gl time bins"""
    return df * np.arange(-(gl // 2), gl - (gl // 2))


def _compute_gaussian_grid_nonumba(x, c, Mr, Msp, tau):
    """Compute the 1D gaussian gridding with Numpy"""
    N = len(x)
    ftau = np.zeros(Mr, dtype=c.dtype)
    hx = 2 * np.pi / Mr
    xmod = x % (2 * np.pi)

    m = 1 + (xmod // hx).astype(int)
    msp = np.arange(-Msp, Msp)[:, np.newaxis]
    mm = m + msp

    E1 = np.exp(-0.25 * (xmod - hx * m) ** 2 / tau)

    # Basically in the following lines we compute
    # this in a tricky way:
    # E2 = np.exp(msp * (xmod - hx * m) * np.pi / (Mr * tau))

    E2 = np.empty((2 * Msp, N), dtype=xmod.dtype)
    E2[Msp] = 1
    E2[Msp + 1:] = np.exp((xmod - hx * m) * np.pi / (Mr * tau))
    E2[Msp + 1:].cumprod(0, out=E2[Msp + 1:])
    E2[Msp - 1::-1] = 1.0 / (E2[Msp + 1] * E2[Msp:])

    E3 = np.exp(-((np.pi * msp / Mr) ** 2) / tau)
    spread = (c * E1) * E2 * E3

    np.add.at(ftau, mm % Mr, spread)

    return ftau


def nuifft(x, y, gl, df=1.0, epsilon=1e-12, exponent="positive"):
    """
    Non-Uniform (inverse) Fast Fourier Transform to avoid linear
    interpolation of interferograms.

    Compute the non-uniform FFT of one-dimensional points x with (complex)
    values y. Result is computed at frequencies (df * gl)
    for integer gl in the range -gl/2 < m < gl/2.
    Uses the fast Gaussian grid algorithm of Greengard & Lee (2004)

    Parameters
    ----------
    x, y : array-like
        real locations x and (complex) values y to be transformed.
    gl, df : int & float
        Parameters specifying the desired frequency grid. Transform will be
        computed at frequencies df * (-(gl//2) + arange(gl))
    epsilon : float
        The desired approximate error for the FFT result. Must be in range
        1E-33 < eps < 1E-1, though be aware that the errors are only well
        calibrated near the range 1E-12 ~ 1E-6.
    exponent : str
        if 'negative', compute the transform with a negative exponent.
        if 'positive', compute the transform with a positive exponent.

    Returns
    -------
    F_k : ndarray
        The complex discrete Fourier transform

    Notes
    -----
    If numba is not installed it's approximately 5x times slower.
    """
    if exponent not in ("positive", "negative"):
        raise ValueError("exponent must be `positive` or `negative`.")

    x = df * np.asarray(x)
    y = np.asarray(y)

    if x.ndim != 1:
        raise ValueError("Expected one-dimensional input arrays")

    if x.shape != y.shape:
        raise ValueError("Array shapes must match")

    gl = int(gl)
    N = len(x)
    k = nuifft_times(gl)

    Mr, Msp, tau = _grid_params(gl, epsilon)

    if _has_numba:
        ftau = _compute_gaussian_grid(x, y, Mr, Msp, tau)
    else:
        warnings.warn("Numba is not available, falling back to slower version.")
        ftau = _compute_gaussian_grid_nonumba(x, y, Mr, Msp, tau)

    if exponent == "negative":
        Ftau = (1 / Mr) * np.fft.fft(ftau)
    else:
        Ftau = np.fft.ifft(ftau)
    Ftau = np.concatenate([Ftau[-(gl // 2):], Ftau[:gl // 2 + gl % 2]])

    return (1 / N) * np.sqrt(np.pi / tau) * np.exp(tau * k ** 2) * Ftau



"""
This file implements the Dataset class with all the functionality
that an interferogram should have in general.
"""
import base64
import warnings
import numbers
import re
from collections.abc import Iterable
from contextlib import suppress
from io import BytesIO
from textwrap import dedent
from math import factorial
import json  # for pretty printing dict
import logging

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.interpolate import interp1d
from jinja2 import Template

from pysprint.config import _get_config_value
from pysprint.core.bases._dataset_base import _DatasetBase, C_LIGHT
from pysprint.core.bases._apply import _DatasetApply
from pysprint.core._evaluate import is_inside
from pysprint.core._evaluate import ifft_method
from pysprint.core._fft_tools import find_center
from pysprint.core.io._parser import _parse_raw
from pysprint.mpl_tools.spp_editor import SPPEditor
from pysprint.mpl_tools.normalize import DraggableEnvelope
from pysprint.utils import MetaData, find_nearest
from pysprint.utils.decorators import inplacify
from pysprint.core._preprocess import (
    savgol,
    find_peak,
    convolution,
    cut_data,
    cwt,
)
from pysprint.utils.exceptions import (
    InterpolationWarning,
    DatasetError,
    PySprintWarning,
)

logger = logging.getLogger(__name__)
FORMAT = "[ %(filename)s:%(lineno)s - %(funcName)20s() ] %(message)s"
logging.basicConfig(format=FORMAT)

__all__ = ["Dataset"]


class Dataset(metaclass=_DatasetBase):
    """
    This class implements all the functionality a dataset
    should have in general.
    """

    meta = MetaData("""Additional info about the dataset""", copy=False)

    def __init__(
            self,
            x,
            y,
            ref=None,
            sam=None,
            meta=None,
            errors="raise",
            callback=None,
            parent=None
    ):
        """
        Base constructor for Dataset.

        Parameters
        ----------
        x : np.ndarray
            The x values.
        y : np.ndarray
            The y values.
        ref : np.ndarray, optional
            The reference arm's spectra.
        sam : np.ndarray, optional
            The sample arm's spectra.
        meta : dict-like
            The dictionary containing further information about the dataset.
            Can be extended, or set to be any valid ~collections.abc.Mapping.
        errors: str, optional
            Whether to raise on missmatching sized data. Must be "raise" or
            "force". If "force" then truncate to the shortest size. Default is
            "raise".
        callback : callable, optional
            The function that notifies parent objects about SPP related
            changes. In most cases the user should leave this empty. The
            default callback is only initialized if this object is constructed
            by the `pysprint.SPPMethod` object.
        parent : any class, optional
            The object which handles the callback function. In most cases
            the user should leave this empty.

        Notes
        -----
        To load in data by files, see the other constructor `parse_raw`.
        """
        super().__init__()

        if errors not in ("raise", "force"):
            raise ValueError("errors must be `raise` or `force`.")

        self.callback = callback or (lambda *args: args)
        self.parent = parent

        self.x = np.array(x, dtype=np.float64)
        self.y = np.array(y, dtype=np.float64)
        if ref is None:
            self.ref = []
        else:
            self.ref = ref
        if sam is None:
            self.sam = []
        else:
            self.sam = sam
        self._is_normalized = False
        if not isinstance(self.x, np.ndarray):
            try:
                self.x = np.array(self.x).astype(float)
            except ValueError:
                raise DatasetError("Invalid type of data")
        if not isinstance(self.y, np.ndarray):
            try:
                self.y = np.array(self.y).astype(float)
            except ValueError:
                raise DatasetError("Invalid type of data")
        if not isinstance(self.ref, np.ndarray):
            try:
                self.ref = np.array(self.ref).astype(float)
            except ValueError:
                pass  # just ignore invalid arms
        if not isinstance(self.sam, np.ndarray):
            try:
                self.sam = np.array(self.sam).astype(float)
            except ValueError:
                pass  # just ignore invalid arms

        if not len(self.x) == len(self.y):
            if errors == 'raise':
                raise ValueError(
                    f"Mismatching data shapes with {self.x.shape} and {self.y.shape}."
                )
            else:
                truncated_shape = min(len(self.x), len(self.y))
                # probably we should cut down the first half
                self.x, self.y = self.x[-truncated_shape:], self.y[-truncated_shape:]

        if len([x for x in (self.ref, self.sam) if len(x) != 0]) == 1:
            warnings.warn(
                "Reference and sample arm should be passed together or neither one.",
                PySprintWarning
            )

        if len(self.ref) == 0 or len(self.sam) == 0:
            self.y_norm = self.y
            self._is_normalized = self._ensure_norm()

        else:
            if not np.all([len(self.sam) == len(self.x), len(self.ref) == len(self.x)]):
                if errors == 'raise':
                    raise ValueError(
                        f"Mismatching data shapes with {self.x.shape}, "
                        f"{self.ref.shape} and {self.sam.shape}."
                    )
                else:
                    truncated_shape = min(len(self.x), len(self.ref), len(self.sam), len(self.y))
                    # same as above..
                    self.ref, self.sam = self.ref[-truncated_shape:], self.sam[-truncated_shape:]

            self.y_norm = (self.y - self.ref - self.sam) / (
                    2 * np.sqrt(self.sam * self.ref)
            )
            self._is_normalized = True

        self.plt = plt
        self.xmin = None
        self.xmax = None
        self.probably_wavelength = None
        self.unit = None
        self._check_domain()

        if meta is not None:
            self.meta = meta

        self._delay = None
        self._positions = None

        nanwarning = np.isnan(self.y_norm).sum()
        infwarning = np.isinf(self.y_norm).sum()
        if nanwarning > 0 or infwarning > 0:
            warnings.warn(
                ("Extreme values encountered during normalization.\n"
                f"Nan values: {nanwarning}\nInf values: {infwarning}"),
                PySprintWarning
            )

        self._dispersion_array = None

    @inplacify
    def chrange(self, current_unit, target_unit="phz"):
        """
        Change the domain range of the dataset.

        Supported units for frequency:
            * PHz
            * THz
            * GHz
        Supported units for wavelength:
            * um
            * nm
            * pm
            * fm

        Parameters
        ----------
        current_unit : str
            The current unit of the domain. Case insensitive.
        target_unit : str, optional
            The target unit. Must be compatible with the currect unit.
            Case insensitive. Default is `phz`.
        """
        current_unit, target_unit = current_unit.lower(), target_unit.lower()
        conversions = {
            "um": {"um": 1, "nm": 1000, "pm": 1E6, "fm": 1E9},
            "nm": {"um": 1 / 1000, "nm": 1, "pm": 1000, "fm": 1E6},
            "pm": {"um": 1 / 1E6, "nm": 1 / 1000, "pm": 1, "fm": 1000},
            "fm": {"um": 1 / 1E9, "nm": 1 / 1E6, "pm": 1 / 1000, "fm": 1},
            "phz": {"phz": 1, "thz": 1000, "ghz": 1E6},
            "thz": {"phz": 1 / 1000, "thz": 1, "ghz": 1000},
            "ghz": {"phz": 1 / 1E6, "thz": 1 / 1000, "ghz": 1}
        }
        try:
            ratio = float(conversions[current_unit][target_unit])
        except KeyError as error:
            raise ValueError("Units are not compatible") from error
        self.x = (self.x * ratio)
        self.unit = self._render_unit(target_unit)
        return self

    def __len__(self):
        return len(self.x)

    @staticmethod
    def _render_unit(unit, mpl=False):
        unit = unit.lower()
        charmap = {
            "um": (r"\mu m", "um"),
            "nm": ("nm", "nm"),
            "pm": ("pm", "pm"),
            "fm": ("fm", "fm"),
            "phz": ("PHz", "PHz"),
            "thz": ("THz", "THz"),
            "ghz": ("GHz", "GHz")
        }
        if mpl:
            return charmap[unit][0]
        return charmap[unit][1]

    @inplacify
    def transform(self, func, axis=None, args=None, kwargs=None):
        """
        Function which enables to apply arbitrary function to the
        dataset.

        Parameters
        ----------
        func : callable
            The function to apply on the dataset.
        axis : int or str, optional
            The axis which is the operation is performed on.
            Must be 'x', 'y', '0' or '1'.
        args : tuple, optional
            Additional arguments to pass to func.
        kwargs : dict, optional
            Additional keyword arguments to pass to func.
        """
        operation = _DatasetApply(
            obj=self, func=func, axis=axis, args=args, kwargs=kwargs
        )
        operation.perform()
        return self

    #  TODO : Rewrite this
    def phase_plot(self, exclude_GD=False):
        """
        Plot the phase if the dispersion is already calculated.

        Parameters
        ----------
        exclude_GD : bool
            Whether to exclude the GD part of the polynomial.
            Default is `False`.
        """
        if not np.all(self._dispersion_array):
            raise ValueError("Dispersion must be calculated before plotting the phase.")

        coefs = np.array(
            [
                self._dispersion_array[i] / factorial(i + 1)
                for i in range(len(self._dispersion_array))
            ]
        )

        if exclude_GD:
            coefs[0] = 0

        phase_poly = np.poly1d(coefs[::-1], r=False)

        self.plt.plot(self.x, phase_poly(self.x))
        self.plt.grid()
        self.plt.ylabel(r"$\Phi\, [rad]$")
        self.plt.xlabel(r"$\omega \,[PHz]$")
        self.plt.show()

    @property
    def delay(self):
        """
        Return the delay value if set.
        """
        return self._delay

    @delay.setter
    def delay(self, value):
        self._delay = value
        try:
            self.callback(self, self.parent)
        except ValueError:
            pass  # delay or position is missing

    @property
    def positions(self):
        """
        Return the SPP position(s) if set.
        """
        return self._positions

    @positions.setter
    def positions(self, value):
        if isinstance(value, numbers.Number):
            if value < np.min(self.x) or value > np.max(self.x):
                raise ValueError(
                    f"Cannot set SPP position to {value} since it's not in the dataset's range."
                )
        # FIXME: maybe we don't need to distinguish between np.ndarray and Iterable
        elif isinstance(value, np.ndarray) or isinstance(value, Iterable):
            for val in value:
                if not isinstance(val, numbers.Number):
                    raise ValueError(
                        f"Expected numeric values, got {type(val)} instead."
                    )
                if val < np.min(self.x) or val > np.max(self.x):
                    raise ValueError(
                        f"Cannot set SPP position to {val} since it's not in the dataset's range."
                    )
        self._positions = value
        try:
            self.callback(self, self.parent)
        except ValueError:
            pass  # delay or position is missing

    def _ensure_norm(self):
        """
        Ensure the interferogram is normalized and only a little part
        which is outlying from the [-1, 1] interval (because of noise).
        """
        try:
            idx = np.where((self.y_norm > 2))
            val = len(idx[0]) / len(self.y_norm)
        except TypeError as e:
            raise DatasetError("Non-numeric values found while reading dataset.") from e
        if val > 0.015:  # this is a custom threshold, which often works..
            return False
        return True

    def scale_up(self):
        """
        If the interferogram is normalized to [0, 1] interval, scale
        up to [-1, 1] with easy algebra.
        """
        self.y_norm = (self.y_norm - 0.5) * 2
        self.y = (self.y - 0.5) * 2

    def GD_lookup(self, reference_point=None, engine="cwt", silent=False, **kwargs):
        """
        Quick GD lookup: it finds extremal points near the
        `reference_point` and returns an average value of 2*pi
        divided by distances between consecutive minimal or maximal values.
        Since it's relying on peak detection, the results may be irrelevant
        in some cases. If the parent class is `~pysprint.CosFitMethod`, then
        it will set the predicted value as initial parameter for fitting.

        Parameters
        ----------
        reference_point : float
            The reference point for the algorithm.
        engine : str, optional
            The backend to use. Must be "cwt", "normal" or "fft".
            "cwt" will use `scipy.signal.find_peaks_cwt` function to
            detect peaks, "normal" will use `scipy.signal.find_peaks`
            to detect peaks. The "fft" engine uses Fourier-transform and
            looks for the outer peak to guess delay value. It's not
            reliable when working with low delay values.
        silent : bool, optional
            Whether to print the results immediately. Default in `False`.
        kwargs : dict, optional
            Additional keyword arguments to pass for peak detection
            algorithms. These are:
            pmin, pmax, threshold, width, floor_thres, etc..
            Most of them are described in the `find_peaks` and
            `find_peaks_cwt` docs.
        """
        precision = _get_config_value("precision")

        if engine not in ("cwt", "normal", "fft"):
            raise ValueError("Engine must be `cwt`, `fft` or `normal`.")

        if reference_point is None and engine != "fft":
            warnings.warn(
                f"Engine `{engine}` isn't available without reference point, falling back to FFT based prediction.",
                PySprintWarning
            )
            engine = "fft"

        if engine == "fft":
            pred, _ = find_center(*ifft_method(self.x, self.y))
            if pred is None:
                if not silent:
                    print("Prediction failed, skipping.")
                return
            print(f"The predicted GD is ± {pred:.{precision}f} fs.")

            if hasattr(self, "params"):
                self.params[3] = pred
            return

        if engine == "cwt":
            widths = kwargs.pop("widths", np.arange(1, 20))
            floor_thres = kwargs.pop("floor_thres", 0.05)
            x_min, _, x_max, _ = self.detect_peak_cwt(
                widths=widths, floor_thres=floor_thres
            )

            # just validation
            _ = kwargs.pop("pmin", 0.1)
            _ = kwargs.pop("pmax", 0.1)
            _ = kwargs.pop("threshold", 0.35)

        else:
            pmin = kwargs.pop("pmin", 0.1)
            pmax = kwargs.pop("pmax", 0.1)
            threshold = kwargs.pop("threshold", 0.35)
            x_min, _, x_max, _ = self.detect_peak(
                pmin=pmin, pmax=pmax, threshold=threshold
            )

            # just validation
            _ = kwargs.pop("widths", np.arange(1, 10))
            _ = kwargs.pop("floor_thres", 0.05)

        if kwargs:
            raise TypeError(f"Invalid argument:{kwargs}")

        try:
            closest_val, idx1 = find_nearest(x_min, reference_point)
            m_closest_val, m_idx1 = find_nearest(x_max, reference_point)
        except (ValueError, IndexError):
            if not silent:
                print("Prediction failed, skipping.. ")
            return
        try:
            truncated = np.delete(x_min, idx1)
            second_closest_val, _ = find_nearest(truncated, reference_point)
        except (IndexError, ValueError):
            if not silent:
                print("Prediction failed, skipping.. ")
            return
        try:
            m_truncated = np.delete(x_max, m_idx1)
            m_second_closest_val, _ = find_nearest(m_truncated, reference_point)
        except (IndexError, ValueError):
            if not silent:
                print("Prediction failed, skipping.. ")
            return

        lowguess = 2 * np.pi / np.abs(closest_val - second_closest_val)
        highguess = 2 * np.pi / np.abs(m_closest_val - m_second_closest_val)

        #  estimate the GD with that
        if hasattr(self, "params"):
            self.params[3] = (lowguess + highguess) / 2

        if not silent:
            print(
                f"The predicted GD is ± {((lowguess + highguess) / 2):.{precision}f} fs"
                f" based on reference point of {reference_point:.{precision}f}."
            )

    def _safe_cast(self):
        """
        Return a copy of key attributes in order to prevent
        inplace modification.
        """
        x, y, ref, sam = (
            np.copy(self.x),
            np.copy(self.y),
            np.copy(self.ref),
            np.copy(self.sam),
        )
        return x, y, ref, sam

    @staticmethod
    def wave2freq(value):
        """Switches a single value between wavelength and angular frequency."""
        return (2 * np.pi * C_LIGHT) / value

    _dispatch = wave2freq.__func__

    @staticmethod
    def freq2wave(value):
        """Switches a single value between angular frequency and wavelength."""
        return Dataset._dispatch(value)

    def _check_domain(self):
        """
        Checks the domain of data just by looking at x axis' minimal value.
        Units are obviously not added yet, we work in nm and PHz...
        """
        try:
            if min(self.x) > 50:
                self.probably_wavelength = True
                self.unit = "nm"
            else:
                self.probably_wavelength = False
                self.unit = "PHz"

        # This is the first function to fail if the user sets up
        # wrong values. Usually..
        except TypeError as error:
            msg = ValueError(
                "The file could not be parsed properly."
            )
            raise msg from error

    @classmethod
    def parse_raw(
        cls,
        filename,
        ref=None,
        sam=None,
        skiprows=0,
        decimal=".",
        sep=None,
        delimiter=None,
        comment=None,
        usecols=None,
        names=None,
        swapaxes=False,
        na_values=None,
        skip_blank_lines=True,
        keep_default_na=False,
        meta_len=1,
        errors="raise",
        callback=None,
        parent=None,
    ):
        """
        Dataset object alternative constructor.
        Helps to load in data just by giving the filenames in
        the target directory.

        Parameters
        ----------
        filename: `str`
            base interferogram
            file generated by the spectrometer
        ref: `str`, optional
            reference arm's spectra
            file generated by the spectrometer
        sam: `str`, optional
            sample arm's spectra
            file generated by the spectrometer
        skiprows: `int`, optional
            Skip rows at the top of the file. Default is `0`.
        decimal: `str`, optional
            Character recognized as decimal separator in the original dataset.
            Often `,` for European data.
            Default is `.`.
        sep: `str`, optional
            The delimiter in the original interferogram file.
            Default is `,`.
        delimiter: `str`, optional
            The delimiter in the original interferogram file.
            This is preferred over the `sep` argument if both given.
            Default is `,`.
        comment: `str`, optional
            Indicates remainder of line should not be parsed. If found at the beginning
            of a line, the line will be ignored altogether. This parameter must be a
            single character. Default is `'#'`.
        usecols: list-like or callable, optional
            If there a multiple columns in the file, use only a subset of columns.
            Default is [0, 1], which will use the first two columns.
        names: array-like, optional
            List of column names to use. Default is ['x', 'y']. Column marked
            with `x` (`y`) will be treated as the x (y) axis. Combined with the
            usecols argument it's possible to select data from a large number of
            columns.
        swapaxes: bool, optional
            Whether to swap x and y values in every parsed file. Default is False.
        na_values: scalar, str, list-like, or dict, optional
            Additional strings to recognize as NA/NaN. If dict passed, specific
            per-column NA values. By default the following values are interpreted as
            NaN: ‘’, ‘#N/A’, ‘#N/A N/A’, ‘#NA’, ‘-1.#IND’,
            ‘-1.#QNAN’, ‘-NaN’, ‘-nan’, ‘1.#IND’, ‘1.#QNAN’,
            ‘<NA>’, ‘N/A’, ‘NA’, ‘NULL’, ‘NaN’, ‘n/a’, ‘nan’, ‘null’.
        skip_blank_lines: bool
            If True, skip over blank lines rather than interpreting as NaN values.
            Default is True.
        keep_default_na: bool
            Whether or not to include the default NaN values when parsing the data.
            Depending on whether na_values is passed in, the behavior changes. Default
            is False. More information available at:
            https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_csv.html
        meta_len: `int`, optional
            The first `n` lines in the original file containing the meta
            information about the dataset. It is parsed to be dict-like.
            If the parsing fails, a new entry will be created in the
            dictionary with key `unparsed`.
            Default is `1`.
        errors: string, optional
            Determines the way how mismatching sized datacolumns behave.
            The default is `raise`, and it will raise on any error.
            If set to `force`, it will truncate every array to have the
            same shape as the shortest column. It truncates from
            the top of the file.
        callback : callable, optional
            The function that notifies parent objects about SPP related
            changes. In most cases the user should leave this empty. The
            default callback is only initialized if this object is constructed
            by the `pysprint.SPPMethod` object.
        parent : any class, optional
            The object which handles the callback function. In most cases
            the user should leave this empty.
        """

        parsed = _parse_raw(
            filename=filename,
            ref=ref,
            sam=sam,
            skiprows=skiprows,
            decimal=decimal,
            sep=sep,
            delimiter=delimiter,
            comment=comment,
            usecols=usecols,
            names=names,
            swapaxes=swapaxes,
            na_values=na_values,
            skip_blank_lines=skip_blank_lines,
            keep_default_na=keep_default_na,
            meta_len=meta_len
        )

        return cls(**parsed, errors=errors, callback=callback, parent=parent)

    def __str__(self):
        _unit = self._render_unit(self.unit)
        precision = _get_config_value("precision")
        string = dedent(
            f"""
        {type(self).__name__}
        ----------
        Parameters
        ----------
        Datapoints: {len(self.x)}
        Predicted domain: {'wavelength' if self.probably_wavelength else 'frequency'}
        Range: from {np.min(self.x):.{precision}f} to {np.max(self.x):.{precision}f} {_unit}
        Normalized: {self._is_normalized}
        Delay value: {str(self._format_delay()) + ' fs' if self._delay is not None else 'Not given'}
        SPP position(s): {str(self._format_positions()) + ' PHz' if np.all(self._positions) else 'Not given'}
        ----------------------------
        Metadata extracted from file
        ----------------------------
        """
        )
        string = re.sub('^\s+', '', string, flags=re.MULTILINE)
        string += json.dumps(self.meta, indent=4, sort_keys=True)
        return string

    def _repr_html_(self):  # TODO: move this to a separate template file
        _unit = self._render_unit(self.unit)
        precision = _get_config_value("precision")
        t = f"""
        <div id="header" class="row" style="height:10%;width:100%;">
        <div style='float:left' class="column">
        <table style="border:1px solid black;float:top;">
        <tbody>
        <tr>
        <td colspan=2 style="text-align:center">
        <font size="5">{type(self).__name__}</font>
        </td>
        </tr>
        <tr>
        <td colspan=2 style="text-align:center">
        <font size="3.5">Parameters</font>
        </td>
        </tr>
        <tr>
        <td style="text-align:center"><b>Datapoints<b></td>
            <td style="text-align:center"> {len(self.x)}</td>
        </tr>
        <tr>
            <td style="text-align:center"><b>Predicted domain<b> </td>
            <td style="text-align:center"> {'wavelength' if self.probably_wavelength else 'frequency'} </td>
        </tr>
        <tr>
        <td style="text-align:center"> <b>Range min</b> </td>
        <td style="text-align:center">{np.min(self.x):.{precision}f} {_unit}</td>
        </tr>
        <tr>
        <td style="text-align:center"> <b>Range max</b> </td>
        <td style="text-align:center">{np.max(self.x):.{precision}f} {_unit}</td>
        </tr>
        <tr>
        <td style="text-align:center"> <b>Normalized</b></td>
        <td style="text-align:center"> {self._is_normalized} </td>
        </tr>
        <tr>
        <td style="text-align:center"><b>Delay value</b></td>
        <td style="text-align:center">{str(self._format_delay()) + ' fs' if self._delay is not None else 'Not given'}</td>
        </tr>
        <tr>
        <td style="text-align:center"><b>SPP position(s)</b></td>
        <td style="text-align:center">{str(self._format_positions()) + ' PHz' if np.all(self._positions) else 'Not given'}</td>
        </tr>
        <tr>
        <td colspan=2 style="text-align:center">
        <font size="3.5">Metadata</font>
        </td>
        </tr>
        """
        jjstring = Template("""
        {% for key, value in meta.items() %}
           <tr>
        <th style="text-align:center"> <b>{{ key }} </b></th>
        <td style="text-align:center"> {{ value }} </td>
           </tr>
        {% endfor %}
            </tbody>
        </table>
        </div>
        <div style='float:leftt' class="column">""")
        rendered_fig = self._render_html_plot()
        return t + jjstring.render(meta=self.meta) + rendered_fig

    def _render_html_plot(self):
        fig, ax = plt.subplots(figsize=(7, 5))
        self.plot(ax=ax)
        plt.close()
        tmpfile = BytesIO()
        fig.savefig(tmpfile, format='png')
        encoded = base64.b64encode(tmpfile.getvalue()).decode('utf-8')
        html_fig = "<img src=\'data:image/png;base64,{}\'>".format(encoded)
        return html_fig

    @property
    def data(self):
        """
        Returns the *current* dataset as `pandas.DataFrame`.
        """
        if self._is_normalized:
            try:
                self._data = pd.DataFrame(
                    {
                        "x": self.x,
                        "y": self.y,
                        "sample": self.sam,
                        "reference": self.ref,
                        "y_normalized": self.y_norm,
                    }
                )
            except ValueError:
                self._data = pd.DataFrame({"x": self.x, "y": self.y})
        else:
            self._data = pd.DataFrame({"x": self.x, "y": self.y})
        return self._data

    # from : https://stackoverflow.com/a/15774013/11751294
    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    @property
    def is_normalized(self):
        """Retuns whether the dataset is normalized."""
        return self._is_normalized

    @inplacify
    def chdomain(self):
        """
        Changes from wavelength [nm] to ang. freq. [PHz]
        domain and vica versa.
        """
        self.x = (2 * np.pi * C_LIGHT) / self.x
        self._check_domain()
        if hasattr(self, "original_x"):
            self.original_x = self.x
        return self

    def detect_peak_cwt(self, widths, floor_thres=0.05):
        """
        Basic algorithm to find extremal points in data
        using ``scipy.signal.find_peaks_cwt``.

        Parameters
        ----------
        widths : np.ndarray
            The widths passed to `find_peaks_cwt`.
        floor_thres : float
            Will be removed.

        Returns
        -------
        xmax : `array-like`
            x coordinates of the maximums
        ymax : `array-like`
            y coordinates of the maximums
        xmin : `array-like`
            x coordinates of the minimums
        ymin : `array-like`
            y coordinates of the minimums
        """
        x, y, ref, sam = self._safe_cast()
        xmax, ymax, xmin, ymin = cwt(
            x, y, ref, sam, widths=widths, floor_thres=floor_thres
        )
        self.xmax = xmax
        self.xmin = xmin
        logger.info(f"{len(xmax)} max values and {len(xmin)} min values found.")
        return xmax, ymax, xmin, ymin

    def savgol_fil(self, window=5, order=3):
        """
        Applies Savitzky-Golay filter on the dataset.

        Parameters
        ----------
        window : int
            Length of the convolutional window for the filter.
            Default is `10`.
        order : int
            Degree of polynomial to fit after the convolution.
            If not odd, it's incremented by 1. Must be lower than window.
            Usually it's a good idea to stay with a low degree, e.g 3 or 5.
            Default is 3.

        Notes
        ------
        If arms were given, it will merge them into the `self.y` and
        `self.y_norm` variables. Also applies a linear interpolation o
        n dataset (and raises warning).
        """
        self.x, self.y_norm = savgol(
            self.x, self.y, self.ref, self.sam, window=window, order=order
        )
        self.y = self.y_norm
        self.ref = []
        self.sam = []
        warnings.warn(
            "Linear interpolation have been applied to data.", InterpolationWarning,
        )

    @inplacify
    def slice(self, start=None, stop=None):
        """
        Cuts the dataset on x axis.

        Parameters
        ----------
        start : float
            Start value of cutting interval.
            Not giving a value will keep the dataset's original minimum value.
            Note that giving `None` will leave original minimum untouched too.
            Default is `None`.
        stop : float
            Stop value of cutting interval.
            Not giving a value will keep the dataset's original maximum value.
            Note that giving `None` will leave original maximum untouched too.
            Default is `None`.

        Notes
        -----
        If arms were given, it will merge them into the `self.y` and
        `self.y_norm` variables. After this operation, the arms' spectra
        cannot be retrieved.
        """
        self.x, self.y_norm = cut_data(
            self.x, self.y, self.ref, self.sam, start=start, stop=stop
        )
        self.ref = []
        self.sam = []
        self.y = self.y_norm
        # Just to make sure it's correctly shaped. Later on we might
        # delete this.
        if hasattr(self, "original_x"):
            self.original_x = self.x
        self._is_normalized = self._ensure_norm()
        return self

    def convolution(self, window_length, std=20):
        """
        Convolve the dataset with a specified Gaussian window.

        Parameters
        ----------
        window_length : int
            Length of the gaussian window.
        std : float
            Standard deviation of the gaussian window.
            Default is `20`.

        Notes
        -----
        If arms were given, it will merge them into the `self.y` and
        `self.y_norm` variables.
        Also applies a linear interpolation on dataset.
        """
        self.x, self.y_norm = convolution(
            self.x, self.y, self.ref, self.sam, window_length, standev=std
        )
        self.ref = []
        self.sam = []
        self.y = self.y_norm
        warnings.warn(
            "Linear interpolation have been applied to data.", InterpolationWarning,
        )

    @inplacify
    def resample(self, N, kind="linear", **kwds):
        """
        Resample the interferogram to have `N` datapoints.

        Parameters
        ----------
        N : int
            The number of datapoints required.
        kind : str, optional
            The type of interpolation to use. Default is `linear`.
        kwds : optional
            Additional keyword argument to pass to `scipy.interpolate.interp1d`.

        Raises
        ------
        PySprintWarning, if trying to subsample to lower `N` datapoints than original.
        """
        f = interp1d(self.x, self.y_norm, kind, **kwds)
        if N < len(self.x):
            N = len(self.x)
            warnings.warn(
                "Trying to resample to lower resolution, keeping shape..", PySprintWarning
            )
        xnew = np.linspace(np.min(self.x), np.max(self.x), N)
        ynew = f(xnew)
        setattr(self, "x", xnew)
        setattr(self, "y_norm", ynew)
        return self

    def detect_peak(self, pmax=0.1, pmin=0.1, threshold=0.1, except_around=None):
        """
        Basic algorithm to find extremal points in data
        using ``scipy.signal.find_peaks``.

        Parameters
        ----------
        pmax : float
            Prominence of maximum points.
            The lower it is, the more peaks will be found.
            Default is `0.1`.
        pmin : float
            Prominence of minimum points.
            The lower it is, the more peaks will be found.
            Default is `0.1`.
        threshold : float
            Sets the minimum distance (measured on y axis) required for a
            point to be accepted as extremal.
            Default is 0.
        except_around : interval (array or tuple),
            Overwrites the threshold to be 0 at the given interval.
            format is `(lower, higher)` or `[lower, higher]`.
            Default is None.

        Returns
        -------
        xmax : `array-like`
            x coordinates of the maximums
        ymax : `array-like`
            y coordinates of the maximums
        xmin : `array-like`
            x coordinates of the minimums
        ymin : `array-like`
            y coordinates of the minimums
        """
        x, y, ref, sam = self._safe_cast()
        xmax, ymax, xmin, ymin = find_peak(
            x,
            y,
            ref,
            sam,
            pro_max=pmax,
            pro_min=pmin,
            threshold=threshold,
            except_around=except_around,
        )
        self.xmax = xmax
        self.xmin = xmin
        logger.info(f"{len(xmax)} max values and {len(xmin)} min values found.")
        return xmax, ymax, xmin, ymin

    def _plot_SPP_if_valid(self, ax=None, **kwargs):
        """
        Mark SPPs on the plot if they are valid.
        """
        if ax is None:
            ax = self.plt
        if isinstance(self.positions, numbers.Number):
            if is_inside(self.positions, self.x):
                x_closest, idx = find_nearest(self.x, self.positions)
                try:
                    ax.plot(x_closest, self.y_norm[idx], **kwargs)
                except (ValueError, TypeError):
                    ax.plot(x_closest, self.y[idx], **kwargs)

        if isinstance(self.positions, np.ndarray) or isinstance(
                self.positions, Iterable
        ):
            if np.array(self.positions).ndim == 0:
                self.positions = np.atleast_1d(self.positions)
            # iterate over 0-d array: need to cast np.atleast_1d
            for i, val in enumerate(self.positions):
                if is_inside(self.positions[i], self.x):
                    x_closest, idx = find_nearest(self.x, self.positions[i])
                    try:
                        ax.plot(x_closest, self.y_norm[idx], **kwargs)
                    except (ValueError, TypeError):
                        ax.plot(x_closest, self.y[idx], **kwargs)

    def _format_delay(self):
        if self.delay is None:
            return ""
        if isinstance(self.delay, np.ndarray):
            if self.delay.size == 0:
                return 0
            delay = np.atleast_1d(self.delay).flatten()
            return delay[0]
        elif isinstance(self.delay, (list, tuple)):
            return self.delay[0]
        elif isinstance(self.delay, numbers.Number):
            return self.delay
        elif isinstance(self.delay, str):
            try:
                delay = float(self.delay)
            except ValueError as e:
                raise TypeError("Delay value not understood.") from e
            return delay
        else:
            raise TypeError("Delay value not understood.")

    def _format_positions(self):
        if self.positions is None:
            return "Not given"
        if isinstance(self.positions, np.ndarray):
            positions = np.atleast_1d(self.positions).flatten()
            return ", ".join(map(str, positions))
        elif isinstance(self.positions, (list, tuple)):
            return ", ".join(map(str, self.positions))
        elif isinstance(self.positions, numbers.Number):
            return self.positions
        elif isinstance(self.positions, str):
            split = self.positions.split(",")
            try:
                positions = [float(p) for p in split]
            except ValueError as e:
                raise TypeError("Delay value not understood.") from e
            return ", ".join(map(str, positions))
        else:
            raise TypeError("Delay value not understood.")

    def _prepare_SPP_data(self):
        pos_x, pos_y = [], []
        if self.positions is not None:
            position = np.array(self.positions, dtype=np.float64).flatten()
            for i, val in enumerate(position):
                if is_inside(position[i], self.x):
                    x_closest, idx = find_nearest(self.x, position[i])
                    try:
                        pos_x.append(x_closest)
                        pos_y.append(self.y_norm[idx])
                    except (ValueError, TypeError):
                        pos_x.append(x_closest)
                        pos_y.append(self.y[idx])
            pos_x = np.array(pos_x)
            pos_y = np.array(pos_y)
        return pos_x, pos_y

    def plot(self, ax=None, title=None, xlim=None, ylim=None, **kwargs):
        """
        Plot the dataset.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            An axis to draw the plot on. If not given, it will plot
            on the last used axis.
        title : str, optional
            The title of the plot.
        xlim : tuple, optional
            The limits of x axis.
        ylim : tuple, optional
            The limits of y axis.
        kwargs : dict, optional
            Additional keyword arguments to pass to plot function.

        Notes
        -----
        If SPP positions are correctly set, it will mark them on plot.
        """
        datacolor = kwargs.pop("color", "red")
        nospp = kwargs.pop("nospp", False)
        _unit = self._render_unit(self.unit, mpl=True)
        xlabel = f"$\lambda\,[{_unit}]$" if self.probably_wavelength else f"$\omega\,[{_unit}]$"
        overwrite = kwargs.pop("overwrite", None)
        if overwrite is not None:
            xlabel = overwrite

        if ax is None:
            ax = self.plt
            self.plt.ylabel("I")
            self.plt.xlabel(xlabel)
            if xlim:
                self.plt.xlim(xlim)
            if ylim:
                self.plt.ylim(ylim)
            if title:
                self.plt.title(title)
        else:
            ax.set(ylabel="I")
            ax.set(xlabel=xlabel)
            if xlim:
                ax.set(xlim=xlim)
            if ylim:
                ax.set(ylim=ylim)
            if title:
                ax.set(title=title)

        if np.iscomplexobj(self.y):
            ax.plot(self.x, np.abs(self.y), color=datacolor, **kwargs)
        else:
            try:
                ax.plot(self.x, self.y_norm, color=datacolor, **kwargs)
            except (ValueError, TypeError):
                ax.plot(self.x, self.y, color=datacolor, **kwargs)
        if not nospp:
            self._plot_SPP_if_valid(ax=ax, color="black", marker="o", markersize=10, label="SPP")

    def show(self):
        """
        Equivalent with plt.show().
        """
        self.plt.show(block=True)

    @inplacify
    def normalize(self, filename=None, smoothing_level=0):
        """
        Normalize the interferogram by finding upper and lower envelope
        on an interactive matplotlib editor. Points can be deleted with
        key `d` and inserted with key `i`. Also points can be dragged
        using the mouse. On complete just close the window. Must be
        called with interactive backend. The best practice is to call
        this function inside `~pysprint.interactive` context manager.

        Parameters
        ----------
        filename : str, optional
            Save the normalized interferogram named by filename in the
            working directory. If not given it will not be saved.
            Default None.
        smoothing_level : int, optional
            The smoothing level used on the dataset before finding the
            envelopes. It applies Savitzky-Golay filter under the hood.
            Default is 0.
        """
        x, y, _, _ = self._safe_cast()
        if smoothing_level != 0:
            x, y = savgol(x, y, [], [], window=smoothing_level)
        _l_env = DraggableEnvelope(x, y, "l")
        y_transform = _l_env.get_data()
        _u_env = DraggableEnvelope(x, y_transform, "u")
        y_final = _u_env.get_data()
        self.y = y_final
        self.y_norm = y_final
        self._is_normalized = True
        self.plt.title("Final")
        self.plot()
        self.show()
        if filename:
            if not filename.endswith(".txt"):
                filename += ".txt"
            np.savetxt(filename, np.column_stack((self.x, self.y)), delimiter=",")
            print(f"Successfully saved as {filename}.")
        return self

    def open_SPP_panel(self, header=None):
        """
        Opens the interactive matplotlib editor for SPP data.
        Use `i` button to add a new point, use `d` key to delete one.
        The delay field is parsed to only get the numeric values.
        Close the window on finish. Must be called with interactive
        backend. The best practice is to call this function inside
        `~pysprint.interactive` context manager.

        Parameters
        ----------
        header : str, optional
            An arbitary string to include as header. This can be
            any attribute's name, or even metadata key.
        """
        if header is not None:
            if isinstance(header, str):
                head = getattr(self, header, None)
                metahead = self.meta.get(header, None)
                info = head or metahead or header
            else:
                info = None
        else:
            info = None
        spp_x, spp_y = self._prepare_SPP_data()
        _spp = SPPEditor(
            self.x, self.y_norm, info=info, x_pos=np.array(spp_x), y_pos=np.array(spp_y)
        )

        textbox = _spp._get_textbox()
        textbox.set_val(self._format_delay())

        _spp._show()

        # We need to split this into separate lines,
        # because empty results are broadcasted twice.
        delay, positions = _spp.get_data()
        self.delay, self.positions = delay, positions

    def emit(self):
        """
        Emit the current SPP data.

        Returns
        -------
        delay : np.ndarray
            The delay value for the current dataset, shaped exactly like
            positions.
        positions : np.ndarray
            The given SPP positions.
        """
        if self.positions is None:
            raise ValueError("SPP positions are missing.")
        if self.delay is None:
            raise ValueError("Delay value is missing.")
        # Important: Use underscored variables to avoid invoking the
        # setter again, which might result in RecursionError and crashes
        # the interpreter.
        if not isinstance(self._positions, np.ndarray):
            self._positions = np.asarray(self.positions)
        if not isinstance(self.delay, np.ndarray):
            self._delay = np.ones_like(self._positions) * self._delay
        return np.atleast_1d(self.delay), np.atleast_1d(self.positions)

    def set_SPP_data(self, delay, positions, force=False):
        """
        Set the SPP data (delay and SPP positions) for the dataset.

        Parameters
        ----------
        delay : float
            The delay value that belongs to the current interferogram.
            Must be given in `fs` units.
        positions : float or iterable
            The SPP positions that belong to the current interferogram.
            Must be float or sequence of floats (tuple, list, np.ndarray, etc.)
        force : bool, optional
            Can be used to set specific SPP positions which are outside of
            the dataset's range. Note that in most cases you should avoid using
            this option. Default is `False`.

        Notes
        -----
        Every position given must be in the current dataset's range, otherwise
        `ValueError` is raised. Be careful to change domain to frequency before
        feeding values into this function.
        """
        if not isinstance(delay, float):
            delay = float(delay)
        delay = np.array(np.ones_like(positions) * delay)
        self.delay = delay
        if force:
            with suppress(ValueError):
                self._positions = positions
        else:
            self.positions = positions



from pysprint import config as cfg


def is_nonnegative_int(value):
    if isinstance(value, int) and value >= 0:
        return True
    raise ValueError


_precision_doc = """
: int
    The precision of floating point numbers in the output.
"""

cfg.register_config_value("precision", 5, doc=_precision_doc, validator=is_nonnegative_int)

_verbosity_doc = """
: int
    The verbosity level of logging outputs when using eager_executor.
"""

cfg.register_config_value("verbosity", 0, doc=_verbosity_doc, validator=is_nonnegative_int)



import numpy as np
import pytest

from pysprint.core._evaluate import (
    is_inside, _split_on_SPP, _build_single_phase_data, min_max_method
)


def test_is_inside():
    x = np.linspace(10, 12, 10)
    assert not is_inside(20, x)
    assert is_inside(11, x)
    assert not is_inside(12, x)
    assert not is_inside(10, x)


def test_splitting_single():
    x = np.array([-10, 2, 3, 4, 5, 6, 9.1])
    splitted = _split_on_SPP(x, 2)
    np.testing.assert_array_equal(splitted[0], np.array([-10.]))
    np.testing.assert_array_equal(splitted[-1], [3, 4, 5, 6, 9.1])
    assert len(splitted) == 2


def test_splitting_multiple():
    x = np.array([-10, 2, 3, 4, 5, 6, 9.1])
    splitted = _split_on_SPP(x, [2, 4])
    np.testing.assert_array_equal(splitted[0], np.array([-10.]))
    np.testing.assert_array_equal(splitted[1], np.array([3]))
    np.testing.assert_array_equal(splitted[-1], [5, 6, 9.1])
    assert len(splitted) == 3


def test_splitting_single_outside():
    x = np.array([-10, 2, 3, 4, 5, 6, 9.1])
    splitted = _split_on_SPP(x, [-20])
    np.testing.assert_array_equal(splitted, [x])
    assert len(splitted) == 1


def test_splitting_multiple_outside():
    x = np.array([-10, 2, 3, 4, 5, 6, 9.1])
    splitted = _split_on_SPP(x, [-20, 30])
    np.testing.assert_array_equal(splitted, [x])
    assert len(splitted) == 1


def test_splitting_single_nearest():
    x = np.array([-10, 2, 3, 4, 5, 6, 9.1])
    splitted = _split_on_SPP(x, 2.2)
    np.testing.assert_array_equal(splitted[0], np.array([-10.]))
    np.testing.assert_array_equal(splitted[-1], [3, 4, 5, 6, 9.1])
    assert len(splitted) == 2


def test_splitting_multiple_nearest():
    x = np.array([-10, 2, 3, 4, 5, 6, 9.1])
    splitted = _split_on_SPP(x, [2.1, 3.9])
    np.testing.assert_array_equal(splitted[0], np.array([-10.]))
    np.testing.assert_array_equal(splitted[1], np.array([3]))
    np.testing.assert_array_equal(splitted[-1], [5, 6, 9.1])
    assert len(splitted) == 3


def test_splitting_multiple_nearest_outside():
    x = np.array([-10, 2, 3, 4, 5, 6, 9.1])
    splitted = _split_on_SPP(x, [-20, 3.9])
    np.testing.assert_array_equal(splitted[0], np.array([-10, 2, 3]))
    np.testing.assert_array_equal(splitted[-1], [5, 6, 9.1])
    assert len(splitted) == 2


def test_build_single_phase_data():
    x = np.arange(11)
    retx, rety = _build_single_phase_data(x)
    np.testing.assert_array_equal(x, retx)
    np.testing.assert_array_equal(np.arange(1, 12) * np.pi, rety)


def test_build_single_phase_data_cb():
    x = np.arange(11)
    retx, rety = _build_single_phase_data(x, SPP_callbacks=[0])
    np.testing.assert_array_equal(np.arange(1, 11), retx)
    np.testing.assert_array_equal(np.arange(1, 11) * -np.pi, rety)


def test_build_single_phase_data_cb2():
    x = np.arange(11)
    retx, rety = _build_single_phase_data(x, SPP_callbacks=[0, 10])
    np.testing.assert_array_equal(np.arange(1, 10), retx)
    np.testing.assert_array_equal(np.arange(1, 10) * -np.pi, rety)


def test_build_single_phase_data_cb3():
    x = np.arange(11)
    retx, rety = _build_single_phase_data(x, SPP_callbacks=[2, 6])
    np.testing.assert_array_equal(np.array([0, 1, 3, 4, 5, 7, 8, 9, 10]), retx)
    np.testing.assert_array_almost_equal(
        np.array([3.141593, 6.283185, 3.141593, 0., -3.141593, 0., 3.141593, 6.283185, 9.424778]), rety
    )


def test_minmax_fail():
    with pytest.raises(TypeError):
        min_max_method(1, 2, 3, 4, SPP_callbacks="sda")


def test_minmax_basic():
    x, y = min_max_method(
        np.arange(10),
        np.arange(10),
        [],
        [],
        ref_point=0,
        minx=np.arange(0, 10)[::2],
        maxx=np.arange(0, 10)[1::2]
    )

    np.testing.assert_array_equal(np.arange(10), x)
    np.testing.assert_array_almost_equal(
        np.array(
            [
                -3.141593,
                -3.141593,
                -6.283185,
                -9.424778,
                -12.566371,
                -15.707963,
                -18.849556,
                -21.991149,
                -25.132741,
                -28.274334
            ]
        ),
        y
    )



import pytest
import numpy as np
import matplotlib.pyplot as plt

from unittest.mock import patch
from pysprint.core.phase import Phase


def test_phase_gd_mode():
    x = np.arange(100)
    y = np.arange(100)

    phase = Phase(x, y, GD_mode=True)
    d, ds, st = phase.fit(reference_point=50, order=2)
    np.testing.assert_array_almost_equal(d, [50, 1])
    assert phase.order == 1
    assert phase.dispersion_order == 2


@patch("matplotlib.pyplot.show")
def test_plots(mck):
    x = np.arange(100)
    y = np.arange(100)

    phase = Phase(x, y)
    phase.plot()
    plt.show()
    mck.assert_called()


def test_errorplot():
    x = np.arange(100)
    y = np.arange(100)

    phase = Phase(x, y)
    phase.fit(2, 2)
    phase.errorplot()
    phase.errorplot(percent=True)


def test_errorplot2():
    x = np.arange(100)
    y = np.arange(100)

    phase = Phase(x, y)
    with pytest.raises(ValueError):
        phase.errors
    with pytest.raises(ValueError):
        phase.errorplot()


def constructor1():
    Phase.from_coeff([1, 2, 3, 4])
    Phase.from_coeff([1, 2, 3, 4], domain=np.arange(5611))


def constructor2():
    Phase.from_disperion_array([1, 2, 3, 4])
    Phase.from_disperion_array([1, 2, 3, 4], domain=np.arange(5611))



import warnings
import logging
from math import factorial
from collections import namedtuple

import numpy as np
from scipy.optimize import curve_fit
from scipy.signal import savgol_filter, medfilt
from scipy.ndimage import gaussian_filter1d

import matplotlib.pyplot as plt

try:
    from lmfit import Model

    _has_lmfit = True
except ImportError:
    _has_lmfit = False

from pysprint.config import _get_config_value
from pysprint.core._functions import _fit_config
from pysprint.core._preprocess import cut_data
from pysprint.utils import pprint_disp
from pysprint.utils import transform_lmfit_params_to_dispersion
from pysprint.utils import transform_cf_params_to_dispersion
from pysprint.utils import _unpack_lmfit
from pysprint.utils import find_nearest
from pysprint.utils import inplacify
from pysprint.utils import NotCalculatedException


logger = logging.getLogger(__name__)
FORMAT = "[ %(filename)s:%(lineno)s - %(funcName)20s() ] %(message)s"
logging.basicConfig(format=FORMAT)


class Phase:
    """
    A class that represents a phase obtained from various
    methods.
    """

    def __init__(self, x, y, GD_mode=False):
        """
        Phase constructor.

        Parameters
        ----------
        x : np.ndarray
            The domain of the phase.
        y : np.ndarray
            The y values of the phase.
        GD_mode : bool, optional
            Whether to treat dataset as a GD graph
            instead of plain phase. Default is False.
        """
        self.x = x
        self.y = y
        self.poly = None
        self.fitted_curve = None
        self.is_dispersion_array = False
        self.is_coeff = False
        self.fitorder = None
        self.GD_mode = GD_mode

        # Make coeffs available after fitting
        self.coef_temp = namedtuple(
            'coef_temp', ['GD', 'GDD', 'TOD', 'FOD', 'QOD', 'SOD']
        )
        self.coef_std_temp = namedtuple(
            'coef_std_temp', ['GD_err', 'GDD_err', 'TOD_err', 'FOD_err', 'QOD_err', 'SOD_err']
        )

        self.coef_array = None
        self.coef_std_array = None

    @property
    def GD(self):
        if self.coef_array is not None:
            return self.coef_array.GD

    @property
    def GDD(self):
        if self.coef_array is not None:
            return self.coef_array.GDD

    @property
    def TOD(self):
        if self.coef_array is not None:
            return self.coef_array.TOD

    @property
    def FOD(self):
        if self.coef_array is not None:
            return self.coef_array.FOD

    @property
    def QOD(self):
        if self.coef_array is not None:
            return self.coef_array.QOD

    @property
    def SOD(self):
        if self.coef_array is not None:
            return self.coef_array.SOD

    @property
    def GD_err(self):
        if self.coef_std_array is not None:
            return self.coef_std_array.GD_err

    @property
    def GDD_err(self):
        if self.coef_std_array is not None:
            return self.coef_std_array.GDD_err

    @property
    def TOD_err(self):
        if self.coef_std_array is not None:
            return self.coef_std_array.TOD_err

    @property
    def FOD_err(self):
        if self.coef_std_array is not None:
            return self.coef_std_array.FOD_err

    @property
    def QOD_err(self):
        if self.coef_std_array is not None:
            return self.coef_std_array.QOD_err

    @property
    def SOD_err(self):
        if self.coef_std_array is not None:
            return self.coef_std_array.SOD_err

    def __call__(self, value):
        if self.poly:
            return self.poly.__call__(value)
        raise NotImplementedError("Before calling, a polinomial must be fitted.")

    @inplacify
    def slice(self, start=None, stop=None):
        """
        Cuts the dataset on x axis.

        Parameters
        ----------
        start : float
            start value of cutting interval
            Not giving a value will keep the dataset's original minimum value.
            Note that giving `None` will leave original minimum untouched too.
            Default is `None`.
        stop : float
            stop value of cutting interval
            Not giving a value will keep the dataset's original maximum value.
            Note that giving `None` will leave original maximum untouched too.
            Default is `None`.
        """
        self.x, self.y = cut_data(self.x, self.y, [], [], start=start, stop=stop)
        return self

    @classmethod
    def from_disperion_array(cls, dispersion_array, domain=None):
        cls.is_dispersion_array = True
        if domain is None:
            x = np.linspace(2, 4, num=2000)
        else:
            x = np.asarray(domain)
        coeffs = [i / factorial(i + 1) for i in dispersion_array]
        cls.poly = np.poly1d(coeffs[::-1])
        return cls(x, cls.poly(x))

    @classmethod
    def from_coeff(cls, GD=0, GDD=0, TOD=0, FOD=0, QOD=0, SOD=0, domain=None):
        if domain is None:
            x = np.linspace(2, 4, num=2000)
        else:
            x = np.asarray(domain)

        cls.is_coeff = True
        cls.poly = np.poly1d([SOD, QOD, FOD, TOD, GDD, GD])
        return cls(x, cls.poly(x))

    def __str__(self):
        if self.poly is not None:
            return self.poly.__str__()
        raise NotImplementedError("Before calling, a polynomial must be fitted.")

    def plot(self, ax=None, marker=None, linestyle=None, **kwargs):
        """
        Plot the phase and the fitted curve (if there's any).

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            An axis to draw the plot on. If not given, it will plot
            of the last used axis.
        kwargs : dict, optional
            Additional keyword arguments to pass to plot function.
        """
        if ax is None:
            ax = plt
            plt.xlabel(r"$\omega\,[PHz]$")
        else:
            ax.set(xlabel=r"$\omega\,[PHz]$")
        if not self.is_dispersion_array or not self.is_coeff:
            # we need to sort them because plots become messy
            # if we keep it unsorted
            idx = np.argsort(self.x)
            x, y = self.x[idx], self.y[idx]
            marker = marker or kwargs.pop("marker", ".")
            linestyle = linestyle or kwargs.pop("linestyle", "None")
            ax.plot(x, y, marker=marker, linestyle=linestyle, **kwargs)
            if self.fitted_curve is not None:
                ax.plot(x, self.fitted_curve[idx], "r--")

    @pprint_disp
    def fit(self, reference_point, order):
        """
        Fit the phase and determine dispersion coefficients.

        Parameters
        ----------

        reference_point : float
            The reference point to use for fitting.
        order : int
            The order of dispersion to look for. Must be in [1, 6].
        """
        return self._fit(reference_point=reference_point, order=order)

    def _fit(self, reference_point, order):
        """
        This is meant to be used privately, when the pprint_disp
        is handled by another function. The `fit` method is for
        public use.
        """
        if self.is_coeff or self.is_dispersion_array:
            warnings.warn("No need to fit another curve.")
            return
        else:
            if self.GD_mode:
                order -= 1
            self.fitorder = order

            _function = _fit_config[order]

            x, y = np.copy(self.x), np.copy(self.y)
            x -= reference_point

            if _has_lmfit:

                fitmodel = Model(_function)
                pars = fitmodel.make_params(**{f"b{i}": 1 for i in range(order + 1)})
                result = fitmodel.fit(y, x=x, params=pars)
            else:
                popt, pcov = curve_fit(_function, x, y, maxfev=8000)

            if _has_lmfit:
                dispersion, dispersion_std = transform_lmfit_params_to_dispersion(
                    *_unpack_lmfit(result.params.items()), drop_first=True, dof=1
                )
                fit_report = result.fit_report()
                self.fitted_curve = result.best_fit
            else:
                dispersion, dispersion_std = transform_cf_params_to_dispersion(
                    popt, drop_first=True, dof=1
                )
                fit_report = (
                    "To display detailed results you must have `lmfit` installed."
                )

                self.fitted_curve = _function(x, *popt)

            if self.GD_mode:
                _, idx = find_nearest(self.x, reference_point)
                dispersion = np.insert(dispersion, 0, self.fitted_curve[idx])
                dispersion_std = np.insert(dispersion_std, 0, 0)

                # The below line must have 7 elements, so slice these redundant coeffs..
                dispersion, dispersion_std = dispersion[:-1], dispersion_std[:-1]

            self.coef_array = self.coef_temp(*dispersion)
            self.coef_std_array = self.coef_std_temp(*dispersion_std)

            return dispersion, dispersion_std, fit_report

    def errorplot(self, ax=None, percent=False, title="Errors", **kwargs):
        """
        Plot the errors of fitting.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            An axis to draw the plot on. If not given, it will plot
            of the last used axis.
        percent : bool, optional
            Whether to plot percentage difference. Default is False.
        title : str, optional
            The title of the plot. Default is "Errors".
        kwargs : dict, optional
            Additional keyword arguments to pass to plot function.
        """
        if ax is None:
            ax = plt.gca()
            plt.xlabel("$\omega\, [PHz]$")
            if percent:
                plt.ylabel("%")
            plt.title(title)
        else:
            ax.set(xlabel="$\omega\, [PHz]$", title=title)
            if percent:
                ax.set(ylabel="%")

        idx = np.argsort(self.x)
        x, y = self.x[idx], self.y[idx]
        if self.fitted_curve is not None:
            if percent:
                ax.plot(x, np.abs((y - self.fitted_curve[idx]) / y) * 100, **kwargs)
            else:
                ax.plot(x, y - self.fitted_curve[idx], **kwargs)
            ax.text(
                0.02,
                0.95,
                f"$R^2 = {self._get_r_squared():.{_get_config_value('precision')}f}$",
                transform=ax.transAxes,
                fontsize="medium",
                bbox=dict(facecolor='white', edgecolor='black', boxstyle='round'),
                zorder=100
            )
            ax.grid()
        else:
            raise NotCalculatedException("Must fit a curve before requesting errors.")

    def flip_around(self, value, side="left"):
        """
        Flip the phase's y values.

        Parameters
        ----------
        value : float
            The x value where to perform flipping.
        side : str, optional
            The side where to flip the y values. Default is "left".
        """
        if side == "left":
            idx = np.where(self.x <= value)[0]
        elif side == "right":
            idx = np.where(self.x >= value)[0]
        else:
            idx = np.array([])

        x_to_flip, y_to_flip = self.x[idx], self.y[idx]

        _, cls_idx = find_nearest(x_to_flip, value)
        logger.info(f"Using {self.x[cls_idx]} instead {value} as flip center.")
        value_base = self.y[cls_idx]
        y_to_flip *= -1
        y_to_flip += 2 * value_base
        self.y[idx] = y_to_flip

    @property
    def errors(self):
        """
        Return the fitting errors as np.ndarray.
        """
        if self.fitted_curve is not None:
            return self.y - self.fitted_curve
        raise ValueError("Must fit a curve before requesting errors.")

    @property
    def order(self):
        """
        Return the order of polynomial which was fitted.
        """
        if self.is_coeff or self.is_dispersion_array:
            self._order = self.poly.order
        else:
            self._order = self.fitorder
        return self._order

    @property
    def dispersion_order(self):
        """
        Return the dispersion order.
        """
        return self.fitorder + 1 if self.GD_mode else self.fitorder

    # TODO : For consistency, this should return a pd.DataFrame.
    # Because we hardwired this into other functions it's not safe to rewrite,
    # but this definitely should be corrected in the future.
    @property
    def data(self):
        """
        Return the data as tuple (x, y).
        """
        return self.x, self.y

    def _get_r_squared(self):
        """
        Calculate the R^2 value.
        """
        if self.fitted_curve is not None:
            residuals = self.y - self.fitted_curve
            ss_res = np.sum(residuals ** 2)
            ss_tot = np.sum((self.y - np.mean(self.y)) ** 2)
            return 1 - (ss_res / ss_tot)
        raise NotCalculatedException

    @property
    def r_squared(self):
        """
        The value of R^2.
        """
        return self._get_r_squared()

    @inplacify
    def remove_range(self, start=None, stop=None):
        """
        Remove a part of the phase for x that satisfies
        start <= x <= stop. Leaving start (stop) as `None`
        will remove from the beginning (end).
        """
        if start is None:
            start = np.min(self.x)
        if stop is None:
            stop = np.max(self.x)

        mask = ((self.x <= stop) & (self.x >= start))
        self.x, self.y = self.x[~mask], self.y[~mask]
        return self

    @inplacify
    def smooth(self, method="gaussian_filter1d", sigma=10, **kwargs):
        """
        Smooth the phase with a piecewise cubic polynomial
        which is twice continuously differentiable.
        """
        if method not in ("savitzky-golay", "convolve", "medfilt", "gaussian_filter1d"):
            raise ValueError(
                "method must be 'savitzky-golay', 'medfilt', 'gaussian_filter1d' or 'convolve'"
            )

        window_length = kwargs.pop("window_lenth", 51)

        if method == "savitzky-golay":
            poly_order = kwargs.pop("poly_order", 3)
            self.y = savgol_filter(self.y, window_length, poly_order, **kwargs)

        elif method == "convolve":
            box = np.ones(window_length) / window_length
            self.y = np.convolve(self.y, box, mode=kwargs.get("mode", "same"))

        elif method == "medfilt":
            self.y = medfilt(self.y, kwargs.get("kernel_size", 5))

        elif method == "gaussian_filter1d":
            sigma = sigma or kwargs.pop("sigma", 10)
            self.y = gaussian_filter1d(self.y, sigma=sigma, **kwargs)

        return self



import os
import collections
import unittest
from unittest.mock import patch

import pytest
import numpy as np
import pandas as pd

from pysprint import Dataset
from pysprint.utils.exceptions import DatasetError, PySprintWarning, InterpolationWarning
from pysprint.config import setting


class TestEvaluate(unittest.TestCase):
    def setUp(self):
        self.x = np.arange(1, 1000, 1)
        self.y = np.sin(self.x)

    def tearDown(self):
        pass

    def test_constructor(self):
        ifg = Dataset(self.x, self.y)
        np.testing.assert_array_equal(ifg.data["x"], self.x)
        np.testing.assert_array_equal(ifg.data["y"], self.y)

    def test_prediction(self):
        ifg = Dataset([150, 200], [3, 4])
        assert ifg.probably_wavelength
        ifg = Dataset([1, 2], [3, 4])
        assert not ifg.probably_wavelength

    def test_dtypes(self):
        ifg = Dataset([15, 4], [3, 4])
        assert type(ifg.x) == np.ndarray
        assert type(ifg.y) == np.ndarray
        with self.assertRaises(ValueError):
            Dataset(["das", 6541], [1, 2])
        with self.assertRaises(ValueError):
            Dataset([2, 6541], ["das", 2])

    def test_safe_casting(self):
        ifg = Dataset([15, 4], [3, 4], [14, 54], [45, 51])
        x, y, ref, sam = ifg._safe_cast()
        assert x is not ifg.x
        np.testing.assert_array_equal(x, ifg.x)
        assert y is not ifg.y
        np.testing.assert_array_equal(y, ifg.y)
        assert ref is not ifg.ref
        np.testing.assert_array_equal(ref, ifg.ref)
        assert sam is not ifg.sam
        np.testing.assert_array_equal(sam, ifg.sam)

    def test_rawparsing(self):
        ifg = Dataset.parse_raw(
            "test_rawparsing.trt", skiprows=8, meta_len=5, decimal=",", sep=";"
        )
        assert issubclass(ifg.meta.__class__, collections.abc.Mapping)
        with self.assertRaises(OSError):
            Dataset.parse_raw(546)

    def test_data(self):
        ifg = Dataset(self.x, self.y)
        assert isinstance(ifg.data, pd.DataFrame)

    @patch("matplotlib.pyplot.show")
    def test_plotting(self, mock_show):
        ifg = Dataset(self.x, self.y)
        ifg.show()
        mock_show.assert_called()

    def test_chdomain(self):
        ifg = Dataset(self.x, self.y)
        before = ifg.x
        ifg.chdomain()
        ifg.chdomain()
        after = ifg.x
        np.testing.assert_array_almost_equal(before, after)

    @pytest.mark.skipif("TF_BUILD" in os.environ, reason="Fails on Azure")
    @patch("matplotlib.pyplot.show")
    def test_normalize(self, mock_show):
        ifg = Dataset(self.x, self.y)
        ifg.normalize()
        mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_sppeditor(self, mock_show):
        ifg = Dataset(self.x, self.y)
        ifg.open_SPP_panel()
        mock_show.assert_called()

    def test_spp_setting(self):
        ifg = Dataset(self.x, self.y)
        delay = 100
        positions = [200, 300]
        ifg.set_SPP_data(delay, positions)
        delay, position = ifg.emit()
        assert len(delay) == len(position)
        np.testing.assert_array_equal(delay, np.array([100, 100]))
        np.testing.assert_array_equal(position, positions)

    def test_slicing_inplace(self):
        ifg = Dataset(self.x, self.y)
        ifg.slice(400, 700)

        assert np.min(ifg.x) > 399
        assert np.max(ifg.x) < 701

    def test_slicing_noinplace(self):
        ifg = Dataset(self.x, self.y)
        new_ifg = ifg.slice(400, 700, inplace=False)

        assert np.min(new_ifg.x) > 399
        assert np.max(new_ifg.x) < 701

        assert not np.min(ifg.x) > 399
        assert not np.max(ifg.x) < 701

    def test_chdomain_noinplace(self):
        ifg = Dataset(self.x, self.y)
        before = ifg.x
        f = ifg.chdomain(inplace=False)
        g = f.chdomain(inplace=False)
        after = g.x
        np.testing.assert_array_almost_equal(before, after)


@pytest.mark.parametrize("pos", [(1, 2), {4, 5}, [4, 5], np.array([1, 4])])
def test_SPP_position_setter_valid_options(pos):
    ifg = Dataset(np.arange(1, 1000, 1), np.sin(np.arange(1, 1000, 1)))
    ifg.positions = pos


@pytest.mark.parametrize("pos", ["string", [1500, 4500], [100, "string"]])
def test_SPP_position_setter_invalid_options(pos):
    ifg = Dataset(np.arange(1, 1000, 1), np.sin(np.arange(1, 1000, 1)))
    with pytest.raises(ValueError):
        ifg.positions = pos


def test_SPP_position_setter_invalid_single():
    ifg = Dataset(np.arange(1, 1000, 1), np.sin(np.arange(1, 1000, 1)))
    with pytest.raises(ValueError):
        ifg.positions = -2000


@patch("matplotlib.pyplot.show")
def test_phase_plot(m):
    fg = Dataset(np.arange(1, 1000, 1), np.sin(np.arange(1, 1000, 1)))
    fg._dispersion_array = np.array([1, 2, 3, 4, 5])
    fg.phase_plot()
    m.assert_called()


@pytest.mark.parametrize("to", ["ghz", "thz"])
def test_chrange_basic(to):
    fg = Dataset(np.arange(1, 1000, 1), np.sin(np.arange(1, 1000, 1)))
    fg.chrange(current_unit="PHz", target_unit=to)
    if to == "thz":
        np.testing.assert_array_almost_equal(np.min(fg.x), 1000)
    elif to == "ghz":
        np.testing.assert_array_almost_equal(np.min(fg.x), 1000000)


@pytest.mark.parametrize("to", ["ghz", "thz"])
def test_chrange_basic_noninplace(to):
    fg = Dataset(np.arange(1, 1000, 1), np.sin(np.arange(1, 1000, 1)))
    ifg = fg.chrange(current_unit="PHz", target_unit=to, inplace=False)
    if to == "thz":
        np.testing.assert_array_almost_equal(np.min(ifg.x), 1000)
    elif to == "ghz":
        np.testing.assert_array_almost_equal(np.min(ifg.x), 1000000)
    np.testing.assert_array_equal(fg.x, np.arange(1, 1000, 1))


def test_chrange_errors():
    fg = Dataset(np.arange(1, 1000, 1), np.sin(np.arange(1, 1000, 1)))
    with pytest.raises(ValueError):
        fg.chrange(current_unit="PHz", target_unit="undefined")


def test_blank_str():
    expected = [
        "Dataset",
        '----------',
        'Parameters',
        '----------',
        'Datapoints: 199',
        'Predicted domain: frequency',
        'Range: from 1.000 to 199.000 PHz',
        'Normalized: True',
        'Delay value: Not given',
        'SPP position(s): Not given',
        '----------------------------',
        'Metadata extracted from file',
        '----------------------------',
        '{}'
    ]
    x_ = np.arange(200)
    y_ = np.linspace(0, 1, 199)

    d = Dataset(x_, y_, errors='force')
    with setting("precision", 3):
        string = d.__str__().split("\n")
    assert string == expected


def test_blank_str_2():
    expected = [
        "Dataset",
        '----------',
        'Parameters',
        '----------',
        'Datapoints: 199',
        'Predicted domain: frequency',
        'Range: from 1.000 to 199.000 PHz',
        'Normalized: True',
        'Delay value: 100 fs',
        'SPP position(s): Not given',
        '----------------------------',
        'Metadata extracted from file',
        '----------------------------',
        '{}'
    ]
    x_ = np.arange(200)
    y_ = np.linspace(0, 1, 199)

    d = Dataset(x_, y_, errors='force')
    d.delay = 100
    with setting("precision", 3):
        string = d.__str__().split("\n")
    assert string == expected


def test_blank_str_3():
    expected = [
        "Dataset",
        '----------',
        'Parameters',
        '----------',
        'Datapoints: 199',
        'Predicted domain: frequency',
        'Range: from 0.000 to 1.000 PHz',
        'Normalized: True',
        'Delay value: 100 fs',
        'SPP position(s): 0.5 PHz',
        '----------------------------',
        'Metadata extracted from file',
        '----------------------------',
        '{}'
    ]
    x_ = np.linspace(0, 1, 199)

    d = Dataset(x_, x_)
    d.delay = 100
    d.positions = 0.5
    with setting("precision", 3):
        string = d.__str__().split("\n")
    assert string == expected


def test_raw_str():
    string = [
        'Dataset',
        '----------',
        'Parameters',
        '----------',
        'Datapoints: 2633',
        'Predicted domain: wavelength',
        'Range: from 360.500 to 1200.250 nm',
        'Normalized: False',
        'Delay value: Not given',
        'SPP position(s): Not given',
        '----------------------------',
        'Metadata extracted from file',
        '----------------------------',
        "{",
        '    "Average": "1 scans",',
        '    "Data measured with spectrometer name": "1107006U1",',
        '    "Integration time": "2,00 ms",',
        '    "Nr of pixels used for smoothing": "0",',
        '    "comment": "m_ifg 8,740",',
        '    "unparsed": [',
        '        [',
        '            "Timestamp [10 microsec ticks]256373440"',
        '        ]',
        '    ]',
        '}'
    ]

    ifg = Dataset.parse_raw("test_rawparsing.trt", skiprows=8, meta_len=5, decimal=",", sep=";")
    with setting("precision", 3):
        actual = ifg.__str__().split("\n")
    assert actual == string


def test_arm_warning():
    with pytest.warns(PySprintWarning):
        Dataset([2], [3], [4])


def test_scale_up():
    x_ = np.linspace(0, 1, 199)

    d = Dataset(x_, x_)
    d.scale_up()

    np.testing.assert_array_almost_equal(d.y, 2 * (x_ - 0.5))
    np.testing.assert_array_almost_equal(d.y_norm, 2 * (x_ - 0.5))


def test_GD_lookup():
    x_ = np.linspace(0, 20, 199)
    d = Dataset(x_, np.sin(x_))
    d.GD_lookup(engine="fft")
    with pytest.warns(PySprintWarning):
        d.GD_lookup()


@pytest.mark.parametrize("val", [-1, 1, 200])
def test_wave2freq(val):
    before = val
    after = Dataset.wave2freq(Dataset.freq2wave(val))
    assert before == after


def test_freq2wave():
    with pytest.raises(ZeroDivisionError):
        Dataset.wave2freq(Dataset.freq2wave(0))


def test_is_normalized():
    x_ = np.linspace(0, 1, 199)
    d = Dataset(x_, x_)
    assert d.is_normalized


def test_is_not_normalized():
    x_ = np.linspace(0, 10, 199)
    d = Dataset(x_, x_)
    assert not d.is_normalized


@pytest.mark.parametrize("func", ["savgol_fil", "convolution"])
def test_specfuncs(func):
    x_ = np.linspace(0, 1, 199)
    d = Dataset(x_, x_)
    with pytest.warns(InterpolationWarning):
        getattr(d, func)(5, 3)
    np.testing.assert_array_equal(d.y, d.y_norm)


def test_resample():
    x_ = np.linspace(0, 1, 199)
    d = Dataset(x_, x_)
    d.resample(200)
    assert len(d) == 200
    d.resample(400, inplace=False)
    assert len(d) == 200
    with pytest.warns(PySprintWarning):
        d.resample(50)


if __name__ == "__main__":
    unittest.main()



import numpy as np

__all__ = ["correct_sign"]


# TODO: rewrite or deprecate this.
def correct_sign(x, flip_increasing=True):
    if not isinstance(x, np.ndarray):
        x = np.asarray(x)
    increasing_idx = np.where(np.diff(x) > 0)
    # inc_len = len(increasing_idx[0])
    decreasing_idx = np.where(np.diff(x) < 0)
    # dec_len = len(decreasing_idx[0])
    # stationary_idx = np.where(np.diff(x) == 0)
    increasing = group_consecutives(increasing_idx)
    decreasing = group_consecutives(decreasing_idx)
    if not (len(increasing) == 1 and len(decreasing) == 1):
        raise ValueError(
            "Values could not be split into two " "strictly monotonic parts safely."
        )
    if flip_increasing:
        x[increasing_idx[0][0]] *= -1
        x[increasing_idx[0] + 1] *= -1
        return x
    x[decreasing_idx[0]] *= -1
    return x


def group_consecutives(arr, step=1):
    return np.split(arr, np.where(np.diff(arr) != step)[0] + 1)



"""
This code is mostly adapted from pandas/pandas/util/_print_versions.py.
pandas is licensed under the BSD 3-Clause "New" or "Revised" License.
See at: pandas.pydata.org
"""
import os
import sys
import platform
import struct
import importlib

__all__ = ["print_info"]

SUPPORTED_VERSIONS = {
    "matplotlib": "2.2.2",
    "scipy": "1.2.0",
    "numpy": "1.16.0",
    "pandas": "0.24.0",
    "lmfit": "1.0.1",
    "numba": "0.46.0",
    "IPython": "7.12.0",
    "pytest": "5.0.1",
    "jinja2": "2.11.1",
    "dask": '2.21.0',
}


def get_system_info():
    uname_result = platform.uname()
    return {
        "python": ".".join(str(j) for j in sys.version_info),
        "python-bits": struct.calcsize("P") * 8,
        "OS": uname_result.system,
        "OS-release": uname_result.release,
        "Version": uname_result.version,
        "machine": uname_result.machine,
        "processor": uname_result.processor,
        "byteorder": sys.byteorder,
    }


def find_version(module):
    version = getattr(module, "__version__", None)
    if version is None:
        raise ImportError(f"Version for {module.__name__} can't be found.")
    return version


def import_optional(name):
    try:
        module = importlib.import_module(name)
    except ImportError:
        return None
    return module


def get_dep_info():
    deps = [
        "pysprint",
        "numpy",
        "scipy",
        "matplotlib",
        "pandas",
        "pytest",
        "lmfit",
        "numba",
        "IPython",
        "jinja2",
        "dask"
    ]
    deps.extend(list(SUPPORTED_VERSIONS))
    result = {}
    for modname in deps:
        mod = import_optional(modname)
        result[modname] = find_version(mod) if mod else None
    return result


def print_info():
    """
    Print all the relevant information about system and dependecies.
    """
    sysinfo = get_system_info()
    deps = get_dep_info()
    n = max(len(x) for x in deps)
    print("\nPYSPRINT ANALYSIS TOOL")
    print("\n        SYSTEM")
    print("----------------------")
    for k, v in sysinfo.items():
        print(f"{k:<{n + 1}}: {v}")
    print("\n      DEPENDENCY")
    print("----------------------")
    for k, v in deps.items():
        print(f"{k:<{n + 1}}: {v}")
    print("\n      ADDITIONAL")
    print("----------------------")
    try:
        ip = __IPYTHON__  # noqa
    except NameError:
        ip = None
    nm = "IPython"
    is_spyder = any("SPYDER" in name for name in os.environ)
    _is_spyder = "Spyder"
    is_conda = os.path.exists(os.path.join(sys.prefix, "conda-meta", "history"))
    _is_conda = "Conda-env"
    print(f"{_is_conda:<{n + 1}}: {is_conda}")
    print(f"{nm:<{n + 1}}: {ip}")
    print(f"{_is_spyder:<{n + 1}}: {is_spyder}")



import os
import contextlib
import warnings
import logging

import numpy as np
from scipy.fftpack import fftshift
import pandas as pd
from scipy.interpolate import interp1d

from pysprint.core.bases.dataset import Dataset
from pysprint.core.bases.algorithms import longest_common_subsequence
from pysprint.core.nufft import nuifft
from pysprint.utils.decorators import inplacify
from pysprint.utils.exceptions import FourierWarning
from pysprint.utils.exceptions import PySprintWarning
from pysprint.utils.exceptions import NotCalculatedException
from pysprint.core._fft_tools import _run
from pysprint.core.phase import Phase
from pysprint.core._evaluate import (
    fft_method,
    cut_gaussian,
    ifft_method,
    gaussian_window,
)

__all__ = ["FFTMethod"]

logger = logging.getLogger(__name__)
FORMAT = "[ %(filename)s:%(lineno)s - %(funcName)20s() ] %(message)s"
logging.basicConfig(format=FORMAT)


class FFTMethod(Dataset):
    """
    Basic interface for the Fourier transform method.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        #  making sure it's not normalized
        if self._is_normalized:
            self.y_norm = self.y
            self._is_normalized = False
        if np.any(self.sam) or np.any(self.ref):
            warnings.warn(
                "This method doesn't require arms' spectra.",
                PySprintWarning
            )
        self.original_x = self.x
        self.at = None
        self.std = None
        self.fwhm = None
        self.window_order = None
        self.phase = None
        self._ifft_called_first = False
        self.nufft_used = False

    @inplacify
    def shift(self, axis="x"):
        """
        Equivalent to `scipy.fftpack.fftshift`, but it's easier to
        use this function instead, because we don't need to explicitly
        call the class' x and y attribute.

        Parameters
        ----------
        axis : str
            Must be 'x', 'y', 'both', 'xy' or 'yx'.
        """
        if axis == "x":
            self.x = fftshift(self.x)
        elif axis == "y":
            self.y = fftshift(self.y)
        elif axis == "both" or axis == "xy" or axis == "yx":
            self.y = fftshift(self.y)
            self.x = fftshift(self.x)
        else:
            raise ValueError("axis should be either `x`, `y` or `both`.")
        return self

    @inplacify
    def ifft(
        self,
        interpolate=True,
        usenifft=False,
        eps=1e-12,
        exponent="positive",
    ):
        """
        Applies inverse Fast Fourier Transfrom to the dataset.

        Parameters
        ----------
        interpolate : bool, default is True -- WILL BE REMOVED
            Whether to apply linear interpolation on the dataset
            before transforming.
        usenifft : bool, optional
            Whether to use non uniform fft. It uses the algorithm
            described in the references. This means the interferogram
            will *not* be linearly interpolated. Default is False.
        eps : float, optional
            The desired approximate error for the non uniform FFT result. Must be
            in range 1E-33 < eps < 1E-1, though be aware that the errors are
            only well calibrated near the range 1E-12 ~ 1E-6. Default is 1E-12.
        exponent : str, optional
            if 'negative', compute the transform with a negative exponent.
            if 'positive', compute the transform with a positive exponent.
            Default is `positive`.

        Notes
        -----
        The basic scheme is ifft -> windowing -> fft, so you should call
        these functions in this order. Otherwise the transforms may be
        inconsistent.

        If numba is not installed the non uniform FTT is approximately
        5x times slower, but still remains comparable to `np.fft.ifft`.

        References
        ----------
        [1] Dutt A., Rokhlin V. : Fast Fourier Transforms for Nonequispaced Data II,
            Applied and Computational Harmonic Analysis
            Volume 2, Issue 1, January 1995, Pages 85-100
            (1995)

        [2] Greengard, Leslie & Lee, June-Yub.: Accelerating the
            Nonuniform Fast Fourier Transform,
            Society for Industrial and Applied Mathematics.
            46. 443-454. 10.1137/S003614450343200X.
            (2004)
        """
        self.nufft_used = usenifft
        self._ifft_called_first = True
        if usenifft:
            x_spaced = np.linspace(self.x[0], self.x[-1], len(self.x))
            timestep = np.diff(x_spaced)[0]
            x_axis = np.fft.fftfreq(len(self.x), d=timestep / (2 * np.pi))
            y_transform = nuifft(
                self.x,
                self.y,
                gl=len(self.x),
                df=(x_axis[1] - x_axis[0]),
                epsilon=eps,
                exponent=exponent,
            )
            self.x, self.y = x_axis, np.fft.fftshift(y_transform)

        else:
            self.x, self.y = ifft_method(self.x, self.y, interpolate=interpolate)
        return self

    @inplacify
    def fft(self):
        """
        Applies fft to the dataset.
        If ifft was not called first, inaccurate results might happen.
        """
        if not self._ifft_called_first:
            warnings.warn(
                "This module is designed to call ifft before fft", FourierWarning
            )
        self.x, self.y = fft_method(self.original_x, self.y)
        return self

    @inplacify
    def window(self, at, fwhm, window_order=6, plot=True):
        """
        Draws a gaussian window on the plot with the desired parameters.
        The maximum value is adjusted for the dataset's maximum value,
        mostly for visibility.

        Parameters
        ----------
        at : float
            The maximum of the gaussian curve.
        fwhm : float
            Full width at half maximum of the gaussian
        window_order : int, optional
            Order of the gaussian curve.
            If not even, it's incremented by 1.
            Default is 6.
        plot : bool, optional
            Whether to immediately show the window with the data.
            Default is `True`.
        """
        self.at = at
        self.fwhm = fwhm
        self.window_order = window_order
        gaussian = gaussian_window(self.x, self.at, self.fwhm, self.window_order)
        self.plt.plot(self.x, gaussian * max(abs(self.y)), "k--")
        if plot:
            self.plot(overwrite="$t\,[fs]$")
            self.show()
        return self

    @inplacify
    def apply_window(self):
        """
        If window function is set, applies window on the dataset.
        """
        self.plt.clf()
        self.plt.cla()
        self.plt.close()
        self.y = cut_gaussian(
            self.x,
            self.y,
            spike=self.at,
            fwhm=self.fwhm,
            win_order=self.window_order,
        )
        return self

    def build_phase(self):
        """
        Retrieve *only the phase* after the transforms. This will
        unwrap the angles and constructs a `~pysprint.core.phase.Phase` object.

        Returns
        -------
        phase : pysprint.core.phase.Phase
            The phase object. See its docstring for more info.
        """
        if self.nufft_used:
            self.shift("y")
        y = np.unwrap(np.angle(self.y), axis=0)
        self.phase = Phase(self.x, y)
        return self.phase  # because of inplace ops. we need to return the phase

    def calculate(self, reference_point, order, show_graph=False):
        """
        FFTMethod's calculate function. It will unwrap the phase by changing
        deltas _between values to 2*pi complement. After that, fit a curve to
        determine dispersion coefficients.

        Parameters
        ----------
        reference_point : float
            The reference point on the x axis.
        order : int
            Polynomial (and maximum dispersion) order to fit. Must be in [1, 5].
        show_graph : bool, optional
            Shows a the final graph of the spectral phase and fitted curve.
            Default is False.

        Returns
        -------
        dispersion : array-like
            The dispersion coefficients in the form of:
            [GD, GDD, TOD, FOD, QOD, SOD]

        dispersion_std : array-like
            Standard deviations due to uncertainty of the fit.
            It is only calculated if lmfit is installed. The form is:
            [GD_std, GDD_std, TOD_std, FOD_std, QOD_std, SOD_std]

        fit_report : str
            If lmfit is available returns the fit report, else returns an
            empty string.

        Notes
        -----
        Decorated with pprint_disp, so the results are immediately
        printed without explicitly saying so.

        Developer-commentary
        --------------------
        Currently the x-axis transformation is sloppy, because we cache the
        original x axis and not transforming it	backwards.
        In addition we need to keep track of interpolation and
        zero-padding too. Currently the transforms are correct only if
        first ifft was used. For now it's doing okay: giving good results.
        For consistency we should still implement that a better way later.
        """
        self.build_phase()
        dispersion, dispersion_std, fit_report = self.phase._fit(
            reference_point=reference_point, order=order
        )
        if show_graph:
            self.phase.plot()

        self._dispersion_array = dispersion
        return -dispersion, dispersion_std, fit_report

    def autorun(
        self,
        reference_point=None,
        order=None,
        *,
        enable_printing=True,
        skip_domain_check=False,
        only_phase=False,
        show_graph=True,
        usenifft=False,
    ):
        """
        Automatically run the Fourier Transfrom based evaluation on the dataset.
        It's not as reliable as I want it to be, so use it carefully. I'm working
        on making it as competent and useful as possible.

        Parameters
        ----------
        reference_point : float, optional
            The reference point on the x axis. If not given, only_phase mode
            will be activated. Default is None.
        order : int, optional
            Polynomial (and maximum dispersion) order to fit. Must be in [1, 6].
            If not given, only_phase mode will be activated. Default is None.
        only_phase : bool, optional
            If True, activate the only_phase mode, which will retrieve the phase
            without fitting a curve, and return a `pysprint.core.Phase.phase` object.
            Default is False (also not giving enough information for curve fitting
            will automatically activate it).
        enable_printing : bool, optional
            If True enable printing the detailed results. Default is True.
        skip_domain_check : bool, optional
            If True skip the interferogram domain check and force the algorithm
            to perform actions without changing domain. If False, check for potential
            wrong domains and change for an appropriate one. Default is False.
        show_graph : bool, optional
            If True show the graph with the phase and the fitted curve, if there is any.
            Default is True.
        usenifft : bool, optional
            If True use the Non Uniform Fast Fourier Transform algorithm. For more details
            see `help(pysprint.FFTMethod.ifft)`. Default is False.

        References
        ----------
        [1] Dutt A., Rokhlin V. : Fast Fourier Transforms for Nonequispaced Data II,
            Applied and Computational Harmonic Analysis
            Volume 2, Issue 1, January 1995, Pages 85-100
            (1995)

        [2] Greengard, Leslie & Lee, June-Yub.: Accelerating the
            Nonuniform Fast Fourier Transform,
            Society for Industrial and Applied Mathematics.
            46. 443-454. 10.1137/S003614450343200X.
            (2004)
        """
        if not reference_point or not order:
            only_phase = True

        if not enable_printing:

            with open(os.devnull, "w") as g, contextlib.redirect_stdout(g):
                _run(
                    self,
                    skip_domain_check=skip_domain_check,
                    show_graph=show_graph,
                    usenifft=usenifft,
                )
            if only_phase:
                y = np.unwrap(np.angle(self.y), axis=0)
                self.phase = Phase(self.x, y)
                return self.phase
            self.calculate(
                reference_point=reference_point, order=order, show_graph=True
            )
        else:
            _run(
                self,
                skip_domain_check=skip_domain_check,
                show_graph=show_graph,
                usenifft=usenifft,
            )
            if only_phase:
                y = np.unwrap(np.angle(self.y), axis=0)
                self.phase = Phase(self.x, y)
                return self.phase
            self.calculate(
                reference_point=reference_point, order=order, show_graph=True
            )

    # TODO: add interpolation
    def get_pulse_shape_from_array(
            self, x_sample, y_sample, truncate=True, tol=None
    ):
        """
        Find out the shape of the pulse in the time domain I(t).

        Parameters
        ----------
        x_sample : np.ndarray
            The x values of the sample arm.
        y_sample : np.ndarray
            The y values of the sample arm.
        truncate : bool, optional
            Whether to truncate the phase and sample spectra
            to the longest_common_subsequence (imeplemented at
            pysprint.core.bases.algorithms). Default is True.
        tol : float or None, optional
            The tolerance which determines how big difference is allowed
            _between x values to interpret them as the same datapoint.
        """
        if self.phase is None:
            raise NotCalculatedException("Must calculate phase first.")
        if not len(y_sample) == len(x_sample):
            raise ValueError("Missmatching shapes.")

        # quick check if we're able to broadcast
        y_sample = np.asarray(y_sample, dtype=float)
        x_phase, y_phase = self.phase.data[0], self.phase.data[1]
        if len(y_sample) != len(self.phase.data[0]):
            if truncate:
                x_sample, y_sample, x_phase, y_phase = longest_common_subsequence(
                    x_sample, y_sample, x_phase, y_phase, tol=tol
                )
                logger.info(
                    f"Shapes were truncated from {np.min(x_sample)} to {np.max(x_sample)} with length {len(x_sample)}."
                )
            else:
                raise ValueError(
                    f"Shapes differ with {len(x_sample)} and {len(self.phase.data[0])}."
                )

        E_field = np.sqrt(y_sample) * np.exp(-1j * y_phase)
        E_pulse = np.abs(np.fft.ifft(E_field)) ** 2

        x_spaced = np.linspace(
            x_phase[0], x_phase[-1], len(x_phase)
        )
        timestep = np.diff(x_spaced)[0]
        x_axis = np.fft.fftfreq(len(x_phase), d=timestep / (2 * np.pi))
        return x_axis, E_pulse

    def get_pulse_shape_from_file(
            self, filename, truncate=True, tol=None, **kwargs
    ):
        """
        Find out the shape of the pulse in the time domain I(t).
        The sample arm's spectra is loaded from file.

        Parameters
        ----------
        filename : str
            The file containing the sample arm's spectra.
        truncate : bool, optional
            Whether to truncate the phase and sample spectra
            to the longest_common_subsequence (imeplemented at
            pysprint.core.bases.algorithms). Default is True.
        tol : float or None, optional
            The tolerance which determines how big difference is allowed
            _between x values to interpret them as the same datapoint.
        kwargs : dict, optional
            The additional keyword arguments for parsing. Same as
            `pysprint.Dataset.parse_raw`. If `chdomain=True`, then
            change the domain after loading.
        """
        if isinstance(filename, str):
            ch = kwargs.pop("chdomain", False)
            df = pd.read_csv(filename, names=["x", "y"], **kwargs)
            x_sample = df["x"].values
            y_sample = df["y"].values
            if ch:
                x_sample = self.wave2freq(x_sample)
            return self.get_pulse_shape_from_array(
                x_sample, y_sample, truncate=truncate, tol=tol
            )

    def errorplot(self, *args, **kwargs):
        """
        Plot the errors of fitting.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            An axis to draw the plot on. If not given, it will plot
            of the last used axis.
        percent : bool, optional
            Whether to plot percentage difference. Default is False.
        title : str, optional
            The title of the plot. Default is "Errors".
        kwargs : dict, optional
            Additional keyword arguments to pass to plot function.
        """
        try:
            getattr(self.phase, "errorplot", None)(*args, **kwargs)
        except TypeError:
            raise NotCalculatedException("Must calculate before plotting errors.")

    @property
    def get_phase(self):
        """
        Return the phase if it is already calculated.
        """
        if self.phase is not None:
            return self.phase
        raise NotCalculatedException("Must retrieve the phase first.")

    @property
    def errors(self):
        """
        Return the fitting errors as np.ndarray.
        """
        errors = getattr(self.phase, "errors", None)
        if errors is not None:
            return errors
        raise NotCalculatedException("Must calculate the fit first.")

    # redefinition to ensure proper attributes are changed

    @inplacify
    def resample(self, N, kind="linear", **kwds):
        """
        Resample the interferogram to have `N` datapoints.

        Parameters
        ----------
        N : int
            The number of datapoints required.
        kind : str, optional
            The type of interpolation to use. Default is `linear`.
        kwds : optional
            Additional keyword argument to pass to `scipy.interpolate.interp1d`.

        Raises
        ------
        PySprintWarning, if trying to subsample to lower `N` datapoints than original.
        """
        f = interp1d(self.x, self.y_norm, kind, **kwds)
        if N < len(self.x):
            N = len(self.x)
            warnings.warn(
                "Trying to resample to lower resolution, keeping shape..", PySprintWarning
            )
        xnew = np.linspace(np.min(self.x), np.max(self.x), N)
        ynew = f(xnew)
        setattr(self, "x", xnew)
        setattr(self, "y_norm", ynew)
        setattr(self, "y", ynew)
        return self



import re

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.widgets import TextBox


class SPPEditor:

    epsilon = 7  # max absolute pixel distance to count as a hit

    def __init__(self, x, y, info=None, x_pos=None, y_pos=None):
        plt.ion()
        self.fig, self.ax = plt.subplots()
        if info is not None:
            if not isinstance(info, str):
                info = str(info)
            self.fig.suptitle(info, fontsize=16)
        self.fig.set_figheight(6)
        self.fig.set_figwidth(10)
        plt.grid()
        plt.subplots_adjust(bottom=0.2)
        self.x = x
        self.y = y
        self._ind = None  # the active point index
        (self.basedata,) = self.ax.plot(self.x, self.y)
        if (x_pos is None) or (y_pos is None):
            x_pos = []
            y_pos = []
        self._setup_positions(x_pos, y_pos)
        self.fig.canvas.mpl_connect(
            "key_press_event", self.key_press_callback
        )
        self.fig.canvas.mpl_connect(
            "button_release_event", self.button_release_callback
        )
        self.axbox = plt.axes([0.1, 0.05, 0.8, 0.1])
        self.text_box = TextBox(
            self.axbox, "Delay [fs]", initial="", color="silver", hovercolor="whitesmoke"
        )
        self.text_box.on_submit(self.submit)
        self.text_box.on_text_change(self.text_change)

    def _setup_positions(self, x_pos, y_pos):
        self.x_pos, self.y_pos = x_pos, y_pos
        (self.points,) = self.ax.plot(self.x_pos, self.y_pos, "ko")

    def _show(self):
        plt.show(block=True)

    def _get_textbox(self):
        return self.text_box

    def _get_ax(self):
        return self.ax

    def submit(self, delay):
        try:
            delay = re.sub(r"[^0-9\.,\-]", "", delay)
            self.delay = float(delay)
        except ValueError:
            pass

    def text_change(self, delay):
        try:
            delay = re.sub(r"[^0-9\.,\-]", "", delay)
            self.delay = float(delay)
        except ValueError:
            pass

    # TODO: Config should decide how to treat missing values.
    def get_data(self):
        positions, _ = self.points.get_data()
        if not hasattr(self, "delay"):
            return np.array([]), np.array([])
        if positions.size != 0:
            self.delay = np.ones_like(positions) * self.delay
        return self.delay, positions

    def button_release_callback(self, event):
        """whenever a mouse button is released"""
        if event.button != 1:
            return
        self._ind = None

    def get_ind_under_point(self, event):
        """
        Get the index of the selected point within the given epsilon tolerance
        """

        # We use the pixel coordinates, because the axes are usually really
        # differently scaled.
        if event.inaxes is None:
            return
        if event.inaxes in [self.ax]:
            try:
                x, y = self.points.get_data()
                xy_pixels = self.ax.transData.transform(np.vstack([x, y]).T)
                xpix, ypix = xy_pixels.T

                # return the index of the point iff within epsilon distance.
                d = np.hypot(xpix - event.x, ypix - event.y)
                (indseq,) = np.nonzero(d == d.min())
                ind = indseq[0]

                if d[ind] >= self.epsilon:
                    ind = None
            except ValueError:
                return
            return ind

    def key_press_callback(self, event):
        """whenever a key is pressed"""
        if not event.inaxes:
            return
        if event.key == "d":
            if event.inaxes in [self.ax]:
                ind = self.get_ind_under_point(event)
            else:
                ind = None
            if ind is not None:
                self.x_pos = np.delete(self.x_pos, ind)
                self.y_pos = np.delete(self.y_pos, ind)
                self.points.set_data(self.x_pos, self.y_pos)

        elif event.key == "i":
            if event.inaxes in [self.ax]:
                self.x_pos = np.append(self.x_pos, event.xdata)
                self.y_pos = np.append(self.y_pos, event.ydata)
                self.points.set_data(self.x_pos, self.y_pos)

        if self.points.stale:
            self.fig.canvas.draw_idle()



from math import factorial

import numpy as np

from pysprint.config import _get_config_value
from pysprint.core.bases.dataset import Dataset
from pysprint.core._optimizer import FitOptimizer
from pysprint.core._evaluate import cff_method
from pysprint.utils import pprint_math_or_default
from pysprint.utils import NotCalculatedException
from pysprint.utils import pad_with_trailing_zeros

__all__ = ["CosFitMethod"]


class CosFitMethod(Dataset):
    """
    Basic interface for the Cosine Function Fit Method.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.params = [1, 1, 1, 1, 1, 1, 1, 1, 1]
        self.fit = None
        self.mt = 8000
        self.f = None
        self.r_squared = None

    def set_max_tries(self, value):
        """
        Overwrite the default scipy maximum try setting to fit the curve.
        """
        self.mt = value

    def adjust_offset(self, value):
        """
        Initial guess for offset.
        """
        self.params[0] = value

    def adjust_amplitude(self, value):
        """
        Initial guess for amplitude.
        """
        self.params[1] = value

    def guess_GD(self, value):
        """
        Initial guess for GD in fs.
        """
        self.params[3] = value

    def guess_GDD(self, value):
        """
        Initial guess for GDD in fs^2.
        """
        self.params[4] = value / 2

    def guess_TOD(self, value):
        """
        Initial guess for TOD in fs^3.
        """
        self.params[5] = value / 6

    def guess_FOD(self, value):
        """
        Initial guess for FOD in fs^4.
        """
        self.params[6] = value / 24

    def guess_QOD(self, value):
        """
        Initial guess for QOD in fs^5.
        """
        self.params[7] = value / 120

    def guess_SOD(self, value):
        """
        Initial guess for SOD in fs^6.
        """
        self.params[8] = value / 720

    def set_max_order(self, order):
        """
        Sets the maximum order of dispersion to look for.
        Should be called after guessing the initial parameters.

        Parameters
        ----------
        order : int
            Maximum order of dispersion to look for. Must be in [1, 6].
        """
        try:
            int(order)
        except ValueError as err:
            raise TypeError(
                "Order should be an in integer from [1, 6]."
            ) from err
        if order > 6 or order < 1:
            raise ValueError(
                "Order should be an in integer from [1, 6]."
            )
        order = 7 - order
        for i in range(1, order):
            self.params[-i] = 0

    def calculate(self, reference_point):
        """
        Cosine fit's calculate function.

        Parameters
        ----------
        reference_point: float
            Reference point on x axis.

        Returns
        -------
        dispersion : array-like
            [GD, GDD, TOD, FOD, QOD, SOD]
        dispersion_std : array-like
            Standard deviations due to uncertainty of the fit.
            They are only calculated if lmfit is installed.
            [GD_std, GDD_std, TOD_std, FOD_std, QOD_std, SOD_std]
        fit_report : str
            Not implemented yet. It returns an empty string for the time being.

        Notes
        ------
        Decorated with pprint_disp, so the results are
        immediately printed without explicitly saying so.
        """
        dispersion, self.fit = cff_method(
            self.x,
            self.y,
            self.ref,
            self.sam,
            ref_point=reference_point,
            p0=self.params,
            maxtries=self.mt,
        )
        precision = _get_config_value("precision")
        self.r_squared = self._get_r_squared()
        pprint_math_or_default(f"R^2 = {self.r_squared:.{precision}f}\n")
        dispersion = pad_with_trailing_zeros(dispersion, 6)
        return (
            dispersion,
            [0, 0, 0, 0, 0, 0],
            "",
        )

    def _get_r_squared(self):
        if self.fit is None:
            raise NotCalculatedException(
                "Must fit a curve before requesting r squared."
            )
        residuals = self.y_norm - self.fit
        ss_res = np.sum(residuals ** 2)
        ss_tot = np.sum((self.y_norm - np.mean(self.y_norm)) ** 2)
        return 1 - (ss_res / ss_tot)

    def plot_result(self):
        """
        If the curve fitting is done, draws the fitted curve on the original
        dataset. Also prints the coeffitient of determination of the
        fit (a.k.a. r^2).
        """
        precision = _get_config_value("precision")
        try:
            self._get_r_squared()
            pprint_math_or_default(f"R^2 = {self.r_squared:.{precision}f}")
        except NotCalculatedException as e:
            raise ValueError("There's nothing to plot.") from e
        if self.fit is not None:
            self.plt.plot(self.x, self.fit, "k--", label="fit", zorder=99)
            self.plt.legend()
            self.plot()
            self.show()
        else:
            self.plot()
            self.show()

    def optimizer(
        self,
        reference_point,
        order=3,
        initial_region_ratio=0.1,
        extend_by=0.1,
        coef_threshold=0.3,
        max_tries=5000,
        show_endpoint=True,
    ):
        """
        Cosine fit optimizer. It's based on adding new terms to fit
        function successively until we reach the maximum order.

        Parameters
        ----------
        reference_point : float
            The reference point on the x axis.
        order : int, optional
            Polynomial (and maximum dispersion) order to fit. Must be in [1, 6].
            Default is 3.
        initial_region_ratio : float, optional
            The initial region in portion of the length of the dataset
            (0.1 will mean 10%, and so on..). Note that the bigger
            resolution the interferogram is, the lower it should be set.
            Default is 0.1. It should not be set too low, because too small
            initial region can result in failure.
        extend_by : float, optional
            The coefficient determining how quickly the region of fit is
            growing. The bigger resolution the interferogram is (or in general
            the higher the dispersion is), the lower it should be set.
            Default is 0.1.
        coef_threshold: float, optional
            The desired R^2 threshold which determines when to expand the region
            of fitting. It's often enough to leave it as is, however if you decide to
            change it, it is highly advised not to set a higher value than 0.7.
            Default is 0.3.
        max_tries : int, optional
            The maximum number of tries to fit a curve before failing.
            Default is 5000.
        show_endpoint : bool, optional
            If True show the fitting results when finished.
            Default is True.

        Notes
        -----
        If the fit fails some parameters must be tweaked in order to
        achieve results. There is a list below with issues,
        its suspected reasons and solutions.

        **SciPy raises OptimizeWarning and the affected area is small
        or not showing any fit**

        Reasons:
        Completely wrong initial GD guess (or lack of guessing).
        Too broad inital region, so that the optimizer cannot find a
        suitable fit.

        This usually happens when the used data is large, or the spectral
        resolution is high.

        Solution:
        Provide better inital guess for GD.
        Lower the inital_region_ratio.

        **SciPy raises OptimizeWarning and the affected area is bigger**

        Reasons: When the optimizer steps up with order it also extends the
        region of fit.

        This error usually present when the region of fit is too quickly
        growing.

        Solution:
        Lower extend_by argument.

        **The optimizer is finished, but wrong fit is produced.**

        Reasons:
        We measure the goodness of fit with r^2 value. To allow this
        optimizer to smoothly find appropriate fits even for noisy datasets
        it's a good practice to keep the r^2 a lower value, such as
        the default 0.3. The way it works is we step up in order of fit
        (until max order) and extend region every time when a fit reaches
        the specified r^2 threshold value. This can be controlled via the
        coef_threshold argument.

        Solution:
        Adjust the coef_threshold value. Note that it's highly
        recommended not to set a higher value than 0.6.
        """
        x, y, ref, sam = self._safe_cast()
        self.f = FitOptimizer(
            x, y, ref, sam, reference_point=reference_point, max_order=order
        )
        self.f.set_initial_region(initial_region_ratio)
        self.f.set_final_guess(
            GD=self.params[3],
            GDD=self.params[4],
            TOD=self.params[5],
            FOD=self.params[6],
            QOD=self.params[7],
            SOD=self.params[8]
        )  # we can pass it higher params safely, they are ignored.
        disp = self.f.run(
            extend_by, coef_threshold, max_tries=max_tries, show_endpoint=show_endpoint,
        )
        disp = disp[3:]
        retval = [disp[i] * factorial(i + 1) for i in range(len(disp))]
        self._dispersion_array = retval
        return retval



import logging
import numbers
from math import factorial
from typing import List, Union, Tuple

import numpy as np
from scipy import fftpack
from scipy.optimize import curve_fit
from scipy.signal import argrelextrema

try:
    from lmfit import Model

    _has_lmfit = True
except ImportError:
    _has_lmfit = False

from pysprint.utils import (
    find_nearest,
    _handle_input,
    _unpack_lmfit,
    _fourier_interpolate,
    transform_cf_params_to_dispersion,
    transform_lmfit_params_to_dispersion,
)

from pysprint.core._functions import _fit_config, _cosfit_config

Num = Union[int, float]
NumericLike = Union[Num, List[Num], np.ndarray]

logger = logging.getLogger(__name__)
FORMAT = "[ %(filename)s:%(lineno)s - %(funcName)20s() ] %(message)s"
logging.basicConfig(format=FORMAT)

__all__ = [
    "min_max_method",
    "spp_method",
    "cff_method",
    "fft_method",
    "cut_gaussian",
    "ifft_method",
    "gaussian_window",
]


def is_inside(value, array):
    """
    Returns whether `value` is strictly inside an array's bounds.
    """
    try:
        return np.any((value < np.max(array)) & (value > np.min(array)))
    except ValueError:
        return False


def _split_on_SPP(a: np.ndarray, val: Union[List, np.ndarray]) -> List[np.ndarray]:
    """
    Split up an array based on value(s).
    """
    if isinstance(val, numbers.Number):
        if is_inside(val, a):
            v, _ = find_nearest(a, val)
            logger.info(f"split value was set to {v} instead of {val}.")
        else:
            logger.info(
                f"{val} is outside of array range, skipping."
            )
            return [a]
        idx = np.where(a != v)[0]
        return np.split(a[idx], np.where(np.diff(idx) != 1)[0] + 1)
    elif isinstance(val, (list, np.ndarray)):
        real_callbacks = []
        for i, v in enumerate(val):
            if not np.any(a == v):
                if is_inside(v, a):
                    value, _ = find_nearest(a, v)
                    real_callbacks.append(value)
                    logger.info(f"{v} was replaced with {value}.")
                else:
                    logger.info(f"{v} was thrown away, not in range..")
            else:
                real_callbacks.append(v)

        idx = np.in1d(a, real_callbacks)
        split_at = a.searchsorted(real_callbacks) - np.arange(0, np.count_nonzero(idx))
        return np.split(a[~idx], split_at)


def _build_single_phase_data(
        x: np.ndarray,
        SPP_callbacks: NumericLike = None,
        flip=False
) -> Tuple[np.ndarray, np.ndarray]:

    y = np.array([])
    lastval = 0

    if SPP_callbacks is not None:
        x = _split_on_SPP(x, SPP_callbacks)
    else:
        x = (x,)
    if flip:
        x.insert(0, [])
    logger.info(f"x was split to {len(x)} pieces (including the flip).")
    for index, i in enumerate(x):
        arr = np.asarray(i)
        if index % 2 == 0:
            y = np.append(y, lastval + np.pi * np.arange(1, len(arr) + 1))
        elif index % 2 == 1:
            y = np.append(y, lastval - np.pi * np.arange(1, len(arr) + 1))
        try:
            lastval = y[-1]
        except IndexError:
            lastval = 0

    return np.concatenate(x), y


def min_max_method(
        x,
        y,
        ref,
        sam,
        ref_point,
        maxx=None,
        minx=None,
        SPP_callbacks=None,
):
    """
    Build the phase from extremal positions and SPP_callbacks.
    """
    x, y = _handle_input(x, y, ref, sam)
    if maxx is None:
        max_ind = argrelextrema(y, np.greater)
        maxx = x[max_ind]
    if minx is None:
        min_ind = argrelextrema(y, np.less)
        minx = x[min_ind]

    _, ref_index = find_nearest(x, ref_point)
    ref_point = x[ref_index]
    logger.info(f"refpoint set to {x[ref_index]} instead of {ref_point}.")

    # subtract the reference point from x axis at extremals
    max_freq = x[ref_index] - maxx
    min_freq = x[ref_index] - minx

    if SPP_callbacks is not None:
        if isinstance(SPP_callbacks, numbers.Number):
            SPP_callbacks -= ref_point
        elif isinstance(SPP_callbacks, (list, np.ndarray)):
            try:
                SPP_callbacks = np.asarray(SPP_callbacks) - ref_point
            except TypeError:
                pass
        else:
            raise TypeError("SPP_callbacks must be list-like, or number.")
        logger.info(f"SPP_callbacks are now {SPP_callbacks}, with ref_point {ref_point}.")

    # find which extremal point is where (relative to reference_point) and order them
    # as they need to be multiplied together with the corresponding order `m`
    neg_freq = np.sort(np.append(max_freq[max_freq < 0], min_freq[min_freq < 0]))[::-1]
    pos_freq = np.sort(np.append(max_freq[max_freq >= 0], min_freq[min_freq >= 0]))

    pos_data_x, pos_data_y = _build_single_phase_data(-pos_freq, SPP_callbacks=SPP_callbacks)

    # if we fail, the whole negative half is empty
    try:
        if np.diff(pos_data_y)[-1] < 0:
            flip = True
            logger.info("Positive side was flipped because the other side is decreasing.")
        else:
            flip = False
    except IndexError:
        flip = False

    neq_data_x, neq_data_y = _build_single_phase_data(-neg_freq, SPP_callbacks=SPP_callbacks, flip=flip)

    x_s = np.insert(neq_data_x, np.searchsorted(neq_data_x, pos_data_x), pos_data_x)
    y_s = np.insert(neq_data_y, np.searchsorted(neq_data_x, pos_data_x), pos_data_y)

    return x_s + ref_point, -y_s + ref_point


def spp_method(delays, omegas, ref_point=0, fit_order=4):
    """
    Calculates the dispersion from SPP's positions and delays.

    Parameters
    ----------
    delays: array-like
        The delay values in fs exactly where omegas taken.
    omegas: array-like
        The angular frequency values where SPP's are located
    ref_point: float
        The reference point in dataset for fitting.
    fit_order: int
        order of polynomial to fit the given data

    Returns
    -------
    omegas: array-like
        x axis data
    delays: array-like
        y axis data
    dispersion: array-like
        [GD, GDD, TOD, FOD, QOD]
    dispersion_std: array-like
        [GD_std, GDD_std, TOD_std, FOD_std, QOD_std]
    bf: array-like
        best fitting curve for plotting
    """
    if fit_order not in range(7):
        raise ValueError("fit order must be in [1, 6]")

    omegas = np.asarray(omegas).astype(np.float64)

    delays = np.asarray(delays).astype(np.float64)

    if not len(delays) == len(omegas):
        raise ValueError(f"data shapes are different: {delays.shape} & {omegas.shape}")

    idx = np.argsort(omegas)
    omegas, delays = omegas[idx], delays[idx]
    omegas -= ref_point

    _function = _fit_config[fit_order]

    if _has_lmfit:
        fitmodel = Model(_function)
        pars = fitmodel.make_params(**{f"b{i}": 1 for i in range(fit_order + 1)})
        result = fitmodel.fit(delays, x=omegas, params=pars)

        dispersion, dispersion_std = transform_lmfit_params_to_dispersion(
            *_unpack_lmfit(result.params.items()), drop_first=False, dof=0
        )
        bf = result.best_fit
    else:
        popt, pcov = curve_fit(_function, omegas, delays, maxfev=8000)
        dispersion, dispersion_std = transform_cf_params_to_dispersion(
            popt, drop_first=False
        )
        bf = _function(omegas, *popt)
    return omegas, delays, -dispersion, dispersion_std, bf


def cff_method(x, y, ref, sam, ref_point=0, p0=None, maxtries=8000):
    """
    Phase modulated cosine function fit method.

    Parameters
    ----------
    x: array-like
        x-axis data
    y: array-like
        y-axis data
    ref, sam: array-like
        the reference and sample arm spectra evaluated at x
    p0: array-like
        the initial parameters for fitting

    Returns
    -------
    dispersion: array-like
        [GD, GDD, TOD, FOD, QOD]
    bf: array-like
        best fitting curve
    """
    if p0 is None:
        p0 = [1, 1, 1, 1, 1, 1, 1, 1, 1]

    x, y = _handle_input(x, y, ref, sam)

    try:
        orderhelper = np.max(np.flatnonzero(p0)) - 2

        p0 = np.trim_zeros(p0, "b")

        _funct = _cosfit_config[orderhelper]

        popt, pcov = curve_fit(_funct, x - ref_point, y, p0, maxfev=maxtries)

        dispersion = np.zeros_like(popt)[:-3]
        for num in range(len(popt) - 3):
            dispersion[num] = popt[num + 3] * factorial(num + 1)
        return dispersion, _funct(x - ref_point, *popt)
    except RuntimeError:
        raise ValueError(
            f"""Max tries ({maxtries}) reached..
                             Parameters could not be estimated."""
        )


def fft_method(x, y):
    """
    Perfoms FFT on data

    Parameters
    ----------
    x: array-like
        the x-axis data
    y: array-like
        the y-axis data

    Returns
    -------
    xf: array-like
        the transformed x data
    yf: array-like
        transformed y data
    """
    yf = fftpack.fft(y)
    xf = np.linspace(x[0], x[-1], len(x))
    return xf, yf


def gaussian_window(t, tau, fwhm, order):
    """
    Returns a simple gaussian window of given parameters evaulated at t.

    Parameters
    ----------
    t: array-like
        input array to perform window on
    tau: float
        center of gaussian window
    fwhm: float
        FWHM of given gaussian
    order: float
        order of gaussian window. If not even it's incremented by 1.

    Returns
    -------
    array : array-like
        nth order gaussian window with params above

    """
    if order % 2 == 1:
        order += 1
    std = fwhm / (2 * (np.log(2) ** (1 / order)))
    return np.exp(-((t - tau) ** order) / (std ** order))


def cut_gaussian(x, y, spike, fwhm, win_order):
    """
    Applies gaussian window with the given params.

    Parameters
    ----------
    x: array-like
        x-axis data
    y: array-like
        y-axis data
    spike: float
        center of gaussian window
    fwhm: float
        Full width at half max
    win_order: int
        The order of gaussian window. Must be even.

    Returns
    -------
    y: array-like
        the windowed y values
    """
    y *= gaussian_window(x, tau=spike, fwhm=fwhm, order=win_order)
    return y


def ifft_method(x, y, interpolate=True):
    """
    Perfoms IFFT on data.

    Parameters
    ----------
    x: array-like
        the x-axis data
    y: array-like
        the y-axis data
    interpolate: bool
        if True perform a linear interpolation on dataset before transforming

    Returns
    -------
    xf: array-like
        the transformed x data
    yf: array-like
        transformed y data
    """
    N = len(x)
    if interpolate:
        x, y = _fourier_interpolate(x, y)
    xf = np.fft.fftfreq(N, d=(x[1] - x[0]) / (2 * np.pi))
    yf = np.fft.ifft(y)
    return xf, yf



import numpy as np
import matplotlib
import matplotlib.pyplot as plt

# ---------------------------MONKEY PATCH-------------------------------------
import matplotlib.cbook as cbook
import matplotlib.widgets as widgets
from matplotlib.backend_tools import ToolBase  # lgtm [py/unused-import]
from matplotlib.backend_managers import ToolManager  # lgtm [py/unused-import]


def tm_init(self, figure=None):
    self._key_press_handler_id = None

    self._tools = {}
    self._keys = {}
    self._toggled = {}
    self._callbacks = cbook.CallbackRegistry()

    # to process keypress event
    self.keypresslock = widgets.LockDraw()
    self.messagelock = widgets.LockDraw()

    self._figure = None
    self.set_figure(figure)


matplotlib.backend_managers.ToolManager.__init__ = tm_init


def tb_init(self, toolmanager, name):
    self._name = name
    self._toolmanager = toolmanager
    self._figure = None


matplotlib.backend_tools.ToolBase.__init__ = tb_init

# ----------------------------------------------------------------------------


from matplotlib.backend_tools import ToolToggleBase
from pysprint.utils import _get_closest


class SelectButton(ToolToggleBase):
    """
    Toggle button on matplotlib toolbar.
    """

    description = "Enable click records"
    default_toggled = True
    default_keymap = "t"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class EditPeak(object):
    """
    This class helps to record and delete peaks from a dataset.
    Right clicks will delete the closest extremal point found
    on the graph, left clicks will add a new point.
    Edits can be saved by just closing the matplotlib window.
    Returns the x coordinates of the selected points.
    Note that this class shouldn't be explicitly called by the user.
    """

    def __init__(self, x, y, x_extremal=None, y_extremal=None):

        # This is here because we do not want other figures
        # to be affected by this.
        plt.ion()
        matplotlib.rcParams["toolbar"] = "toolmanager"

        self.figure, self.ax = plt.subplots()
        self.cid = None
        self.x = x
        self.y = y
        plt.plot(self.x, self.y, "r")
        self.x_extremal = x_extremal
        self.y_extremal = y_extremal
        if not len(self.x_extremal) == len(self.y_extremal):
            raise ValueError("Data shapes are different")
        self.press()
        (self.lins,) = self.ax.plot(
            self.x_extremal, self.y_extremal, "ko", markersize=6, zorder=99
        )
        self.ax.grid(alpha=0.7)
        # adding the button to navigation toolbar
        tm = self.figure.canvas.manager.toolmanager
        tm.add_tool("Toggle recording", SelectButton)
        self.figure.canvas.manager.toolbar.add_tool(
            tm.get_tool("Toggle recording"), "toolgroup"
        )
        self.my_select_button = tm.get_tool("Toggle recording")
        plt.sca(self.ax)
        plt.show(block=True)

    def on_clicked(self, event):
        """ Function to record and discard points on plot."""
        ix, iy = event.xdata, event.ydata
        if event.inaxes is None:
            return
        if self.my_select_button.toggled:
            if event.key == "d":
                ix, iy, idx = _get_closest(ix, iy, self.x_extremal, self.y_extremal)
                self.x_extremal = np.delete(self.x_extremal, idx)
                self.y_extremal = np.delete(self.y_extremal, idx)
            elif event.key == "i":
                ix, iy, idx = _get_closest(ix, iy, self.x, self.y)
                self.x_extremal = np.append(self.x_extremal, ix)
                self.y_extremal = np.append(self.y_extremal, iy)
            self.lins.set_data(self.x_extremal, self.y_extremal)
            plt.draw()
            return

    def press(self):
        """Usual function to connect matplotlib.."""
        self.cid = self.figure.canvas.mpl_connect("key_press_event", self.on_clicked)

    def release(self):
        """ On release functionality. It's never called but we will
        need this later on.."""
        self.figure.canvas.mpl_disconnect(self.cid)

    @property
    def get_dat(self):
        """ Returns the the selected points."""
        return self.lins.get_data()



from unittest.mock import patch
import importlib

import pytest
import numpy as np

from pysprint import Generator, WFTMethod
from pysprint.utils import NotCalculatedException


@pytest.mark.slow
@patch('matplotlib.pyplot.show')
def test_basic(mck):
    g = Generator(
        1,
        3,
        2,
        3000,
        GDD=400,
        TOD=4000,
        FOD=4000,
        QOD=50000,
        pulse_width=5,
        resolution=0.01
    )

    g.generate()

    f = WFTMethod(*g.data)
    f.add_window_linspace(1.25, 2.75, 350, fwhm=0.017)

    d, _, _ = f.calculate(
        reference_point=2, order=5, fastmath=False, silent=True, parallel=False
    )

    np.testing.assert_array_almost_equal(
        d, [2999.15, 399.94, 3998.03, 3991.45, 49894.96], decimal=1
    )

    f.heatmap()
    f.show()
    mck.assert_called()


def test_window_basic():
    w = WFTMethod(np.arange(100), np.arange(100))
    w.add_window(20, fwhm=1)
    with pytest.raises(TypeError):
        w.add_window(515)
    w.remove_window_at(20)
    w.add_window_linspace(1, 20, 19, fwhm=1)
    w.remove_window_interval(1, 20)
    assert len(w.windows) == 0


@patch('matplotlib.pyplot.show')
def test_window_generic(mock_show):
    w = WFTMethod(np.arange(100), np.arange(100))
    w.add_window_generic(np.random.normal(50, 10, size=50), fwhm=50)
    assert len(w.windows) == 50
    w.view_windows()
    w.reset_state()
    assert len(w.windows) == 0
    with pytest.raises(NotCalculatedException):
        w.heatmap()
    with pytest.raises(NotCalculatedException):
        w.get_GD()


@pytest.mark.skipif(
    importlib.util.find_spec('dask') is None,
    reason="dask is required"
)
@pytest.mark.slow
def test_basic_parallel():
    g = Generator(
        1,
        3,
        2,
        3000,
        GDD=400,
        TOD=4000,
        FOD=4000,
        QOD=50000,
        pulse_width=5,
        resolution=0.01
    )

    g.generate()

    f = WFTMethod(*g.data)
    f.add_window_linspace(1.25, 2.75, 350, fwhm=0.017)

    d, _, _ = f.calculate(
        reference_point=2, order=5, fastmath=False, parallel=True, silent=True
    )

    np.testing.assert_array_almost_equal(
        d, [2999.16249, 399.94162, 3998.03069, 3991.45663, 49894.96710], decimal=1
    )



from collections.abc import Mapping
from collections import OrderedDict
from copy import deepcopy

__all__ = ["MetaData"]


# This class is adapted from AstroPy licensed under
# the BSD 3-Clause "New" or "Revised" License.
# source: https://github.com/astropy/astropy/blob/master/astropy/utils/metadata.py
class MetaData:
    """
    A class to store additional meta property.
    This can be set to any valid ~collections.abc.Mapping.

    Parameters
    ----------
    doc : `str`, optional
        Documentation for the attribute of the class.
        Default is ``""``.

    copy : `bool`, optional
        If ``True`` the the value is deepcopied before setting, otherwise it
        is saved as reference.
        Default is ``True``.

    """

    def __init__(self, doc="", copy=True):
        self.__doc__ = doc
        self.copy = copy

    def __get__(self, instance, owner):
        if instance is None:
            return self
        if not hasattr(instance, "_meta"):
            instance._meta = OrderedDict()
        return instance._meta

    def __set__(self, instance, value):
        if value is None:
            instance._meta = OrderedDict()
        else:
            if isinstance(value, Mapping):
                if self.copy:
                    instance._meta = deepcopy(value)
                else:
                    instance._meta = value
            else:
                raise TypeError("meta attribute must be dict-like")



import os
import warnings
from functools import lru_cache

import numpy as np
import matplotlib.pyplot as plt  # noqa
import matplotlib.cbook as cbook

from pysprint.core.bases.dataset import Dataset
from pysprint.core.bases._dataset_base import _DatasetBase
from pysprint.core.phase import Phase
from pysprint.utils.exceptions import DatasetError
from pysprint.core.callbacks import defaultcallback

__all__ = ["SPPMethod"]

warnings.simplefilter("ignore", category=cbook.MatplotlibDeprecationWarning)


class SPPMethod(metaclass=_DatasetBase):
    """
    Interface for Stationary Phase Point Method.
    """

    def __init__(self, ifg_names, sam_names=None, ref_names=None, errors="raise", **kwargs):
        """
        SPPMethod constructor.

        Parameters
        ----------
        ifg_names : list
            The list containing the filenames of the interferograms.
        sam_names : list, optional
            The list containing the filenames of the sample arm's spectra.
        ref_names : list, optinal
            The list containing the filenames of the reference arm's spectra.
        kwargs :
            Additional keyword arguments to pass to `parse_raw` function.
        """
        if errors not in ("raise", "ignore"):
            raise ValueError("errors must be `raise` or `ignore`.")

        self.ifg_names = ifg_names

        if sam_names:
            self.sam_names = sam_names
        else:
            self.sam_names = None
        if ref_names:
            self.ref_names = ref_names
        else:
            self.ref_names = None

        if errors == "raise":
            self._validate()
            if self.sam_names:
                if not len(self.ifg_names) == len(self.sam_names):
                    raise DatasetError(
                        "Missmatching length of files. Use None if a file is missing."
                    )
            if self.ref_names:
                if not len(self.ifg_names) == len(self.ref_names):
                    raise DatasetError(
                        "Missmatching length of files. Use None if a file is missing."
                    )

        self.skiprows = kwargs.pop("skiprows", 0)
        self.decimal = kwargs.pop("decimal", ",")
        self.sep = kwargs.pop("sep", ";")
        self.meta_len = kwargs.pop("meta_len", 1)
        self.cb = kwargs.pop("callback", defaultcallback)
        self.delimiter = kwargs.pop("delimiter", None)
        self.comment = kwargs.pop("comment", None)
        self.usecols = kwargs.pop("usecols", None)
        self.names = kwargs.pop("names", None)
        self.swapaxes = kwargs.pop("swapaxes", False)
        self.na_values = kwargs.pop("na_values", None)
        self.skip_blank_lines = kwargs.pop("skip_blank_lines", True)
        self.keep_default_na = kwargs.pop("keep_default_na", False)

        if kwargs:
            raise TypeError(f"invalid keyword argument:{kwargs}")

        self.load_dict = {
            "skiprows": self.skiprows,
            "decimal": self.decimal,
            "sep": self.sep,
            "meta_len": self.meta_len,
            "callback": self.cb,
            "delimiter": self.delimiter,
            "comment": self.comment,
            "usecols": self.usecols,
            "names": self.names,
            "swapaxes": self.swapaxes,
            "na_values": self.na_values,
            "skip_blank_lines": self.skip_blank_lines,
            "keep_default_na": self.keep_default_na
        }

        self._container = {}
        self._info = f"Progress: {len(self._container)}/{len(self)}"
        self.GD = None

    def _collect(self):

        # Maybe the dictionary struct can be dropped at this point..
        local_delays = {}
        local_positions = {}

        for idx, (delay, position) in enumerate(self._container.values()):
            if idx != 0 and delay.size > 0 and delay.flat[0] in np.concatenate(
                    [a.ravel() for a in local_delays.values()]
            ):
                raise ValueError(
                    f"Duplicated delay values found. Delay {delay.flat[0]} fs was previously seen."
                )
            local_delays[idx] = delay
            local_positions[idx] = position

        delays = np.concatenate([a.ravel() for a in local_delays.values()])
        positions = np.concatenate([a.ravel() for a in local_positions.values()])

        return delays, positions

    def append(self, newifg, newsam=None, newref=None):
        """
        Append a new interferogram to the object.
        """
        # ensure padding before trying to append, and also
        # we better prevent infinite loop

        # TODO
        self.ifg_names.append(newifg)
        if newsam is not None:
            if self.sam_names is not None:
                if len(self.ifg_names) > len(self.sam_names):
                    while len(self.ifg_names) != len(self.sam_names):
                        self.sam_names.append(None)
                self.sam_names.append(newsam)
        if newref is not None:
            if self.ref_names is not None:
                if len(self.ifg_names) > len(self.ref_names):
                    while len(self.ifg_names) != len(self.ref_names):
                        self.ref_names.append(None)
                self.ref_names.append(newref)

    @staticmethod
    def calculate_from_ifg(ifg_list, reference_point, order, show_graph=False):
        """
        Collect SPP data from a list of `pysprint.Dataset` or child objects
        and evaluate them.

        Parameters
        ----------
        ifg_list : list
            The list containing the interferograms. All member should be
            `pysprint.Dataset` or child class type, otherwise TypeError is raised.
        reference_point : float
            The reference point on the x axis.
        order : int
            Maximum dispersion order to look for. Must be in [2, 6].
        show_graph : bool, optional
            Shows a the final graph of the spectral phase and fitted curve.
            Default is False.

        Returns
        -------
        dispersion : array-like
            The dispersion coefficients in the form of:
            [GD, GDD, TOD, FOD, QOD, SOD]
        dispersion_std : array-like
            Standard deviations due to uncertainty of the fit.
            It is only calculated if lmfit is installed. The form is:
            [GD_std, GDD_std, TOD_std, FOD_std, QOD_std, SOD_std]
        fit_report : str
            If lmfit is available returns the fit report, else returns an
            empty string.
        """
        for ifg in ifg_list:
            if not isinstance(ifg, Dataset):
                raise TypeError("pysprint.Dataset objects are expected.")
        if order == 1:
            raise ValueError(
                "Order should be greater than 1. Cannot fit constant function to data."
            )

        local_delays = {}
        local_positions = {}

        for idx, ifg in enumerate(ifg_list):
            delay, position = ifg.emit()
            if idx != 0 and delay.size > 0 and delay.flat[0] in np.concatenate(
                    [a.ravel() for a in local_delays.values()]
            ):
                raise ValueError(
                    f"Duplicated delay values found. Delay {delay.flat[0]} fs was previously seen."
                )
            local_delays[idx] = delay
            local_positions[idx] = position
        delays = np.concatenate([a.ravel() for a in local_delays.values()])
        positions = np.concatenate([a.ravel() for a in local_positions.values()])
        GD = Phase(positions, delays, GD_mode=True)
        d, ds, s = GD._fit(reference_point=reference_point, order=order)

        if show_graph:
            GD.plot()

        return d, ds, s

    def __len__(self):
        return len(self.ifg_names)

    # TODO
    def __str__(self):
        return f"{type(self).__name__}\nInterferogram count : {len(self)}"

    def _repr_html_(self):
        alive = [i for i in Dataset._get_instances() if i.parent == self]
        s = f"""
        <table style="border:1px solid black;float:top;">
        <tbody>
        <tr>
        <td colspan=2 style="text-align:center">
        <font size="5">{type(self).__name__}</font>
        </td>
        </tr>
        <tr>
        <td style="text-align:center"><b>Interferograms accumulated<b></td>
            <td style="text-align:center"> {len(self)}</td>
        </tr>
        <tr>
        <td style="text-align:center"><b>Interferograms cached<b></td>
            <td style="text-align:center"> {len(alive)}</td>
        </tr>
        <tr>
        <td style="text-align:center"><b>Eagerly calculating<b></td>
            <td style="text-align:center"> {self.is_eager}</td>
        </tr>
        <tr>
        <td style="text-align:center"><b>Data recorded from<b></td>
            <td style="text-align:center"> {len(self._container)}</td>
        </tr>
        </table>
        """
        return s

    @lru_cache(500)
    def __getitem__(self, key):
        if isinstance(key, slice):
            raise TypeError("Slices are not acceptable.")
        try:
            dataframe = Dataset.parse_raw(
                self.ifg_names[key],
                self.sam_names[key],
                self.ref_names[key],
                **self.load_dict,
                parent=self
            )
        except (TypeError, ValueError):
            dataframe = Dataset.parse_raw(
                self.ifg_names[key],
                **self.load_dict,
                parent=self
            )
        return dataframe

    def _validate(self):
        for filename in self.ifg_names:
            if filename is not None and not os.path.exists(filename):
                raise FileNotFoundError(f"""File named '{filename}' is not found.""")
        if self.sam_names:
            for sam in self.sam_names:
                if sam is not None and not os.path.exists(sam):
                    raise FileNotFoundError(f"""File named '{sam}' is not found.""")
        if self.ref_names:
            for ref in self.ref_names:
                if ref is not None and not os.path.exists(ref):
                    raise FileNotFoundError(f"""File named '{ref}' is not found.""")

    def flush(self):
        """
        Reset the state of recorded delays and positions, even on
        active objects that have been constructed on the runtime.
        """
        self._container = {}
        for ifg in self:
            ifg._delay = None
            ifg.delay = None
            ifg._positions = None
            ifg.positions = None

    def save_data(self, filename):
        """
        Save the currectly stored SPP data.

        Parameters
        ----------
        filename : str
            The filename to save as. If not ends with ".txt" it's
            appended by default.
        """
        if not filename.endswith(".txt"):
            filename += ".txt"
        delays, positions = self._collect()
        np.savetxt(
            f"{filename}", np.column_stack((positions, delays)), delimiter=",",
        )

    @staticmethod
    def calculate_from_raw(omegas, delays, reference_point, order, show_graph=False):
        """
        Calculate the dispersion from matching pairs of delays and SPP positions.

        Parameters
        ----------
        omegas : np.ndarray
            The SPP positions.
        delays : np.ndarray
            The delay values in fs.
        reference_point : float
            The reference point on the x axis.
        order : int
            Maximum dispersion order to look for. Must be in [2, 6].
        show_graph : bool, optional
            Whether to show the fitting.

        Returns
        -------
        dispersion : array-like
            The dispersion coefficients in the form of:
            [GD, GDD, TOD, FOD, QOD, SOD]
        dispersion_std : array-like
            Standard deviations due to uncertainty of the fit.
            It is only calculated if lmfit is installed. The form is:
            [GD_std, GDD_std, TOD_std, FOD_std, QOD_std, SOD_std]
        fit_report : str
            If lmfit is available returns the fit report, else returns an
            empty string.
        """
        if order == 1:
            raise ValueError(
                "Order should be greater than 1. Cannot fit constant function to data."
            )

        GD = Phase(omegas, delays, GD_mode=True)
        d, ds, s = GD._fit(reference_point=reference_point, order=order)

        if show_graph:
            GD.plot()

        return d, ds, s

    def build_GD(self):
        """
        Build the GD and return it. It will rebuild every time
        this function is called.
        """
        delays, positions = self._collect()

        self.GD = Phase(positions, delays, GD_mode=True)
        return self.GD

    def calculate(self, reference_point, order, show_graph=False):
        """
        This function should be used after setting the SPP data in
        the interactive matplotlib editor or other way.

        Parameters
        ----------
        reference_point : float
            The reference point on the x axis.
        order : int, optional
            Maximum dispersion order to look for. Must be in [2, 6].
        show_graph : bool, optional
            Shows a the final graph of the spectral phase and fitted curve.
            Default is False.

        Returns
        -------
        dispersion : array-like
            The dispersion coefficients in the form of:
            [GD, GDD, TOD, FOD, QOD, SOD]
        dispersion_std : array-like
            Standard deviations due to uncertainty of the fit.
            It is only calculated if lmfit is installed. The form is:
            [GD_std, GDD_std, TOD_std, FOD_std, QOD_std, SOD_std]
        fit_report : str
            If lmfit is available returns the fit report, else returns an
            empty string.
        """
        if order == 1:
            raise ValueError(
                "Order should be greater than 1. Cannot fit constant function to data."
            )
        self.build_GD()
        d, ds, s = self.GD._fit(reference_point=reference_point, order=order)

        if show_graph:
            self.GD.plot()

        return d, ds, s

    @property
    def info(self):
        """
        Return how many interferograms were processed.
        """
        self._info = f"Progress: {len(self._container)}/{len(self)}"
        return self._info

    @property
    def is_eager(self):
        """
        Returns if eager execution is enabled.
        """
        # TODO
        if self.cb.__name__ == "inner":
            return True
        return False
